#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_allclose

from fastoad.exceptions import FastUnexpectedKeywordArgument
from ..flight_point import FlightPoint


def test_FlightPoint():

    # Init with kwargs, with one attribute initialized #############################################
    fp1 = FlightPoint(mass=50000.0)
    assert fp1.mass == 50000.0
    assert fp1.mass == fp1["mass"]
    fp1.mass = 70000.0
    assert fp1["mass"] == 70000.0

    # Non initialized but known attributes are initialized to None
    for label in FlightPoint.get_attribute_keys():
        if label != "mass":
            assert getattr(fp1, label) is None

    # Init with dictionary, with all attributes initialized ########################################
    test_values = {
        key: value
        for key, value in zip(
            FlightPoint.get_attribute_keys(), range(len(FlightPoint.get_attribute_keys()))
        )
    }
    fp2 = FlightPoint(test_values)
    for label in FlightPoint.get_attribute_keys():
        assert getattr(fp2, label) == test_values[label]
        assert getattr(fp2, label) == fp2[label]
        new_value = test_values[label] * 100
        setattr(fp2, label, new_value)
        assert fp2[label] == new_value

    # Init with unknown attribute ##################################################################
    with pytest.raises(FastUnexpectedKeywordArgument):
        _ = FlightPoint(unknown=0)

    # Unknown dictionary keys are allowed, though
    _ = FlightPoint({"unknown": 0})

    # FlightPoint to/from pandas DataFrame #########################################################
    assert fp1 == FlightPoint(pd.DataFrame([fp1]).iloc[0])

    df = pd.DataFrame([fp1, fp2])
    for label in FlightPoint.get_attribute_keys():
        assert_allclose(df[label], [fp1.get(label, np.nan), fp2.get(label, np.nan)])

    fp2bis = FlightPoint(df.iloc[-1])
    assert fp2 == fp2bis

    fp1bis = FlightPoint(df.iloc[0])
    assert fp1 == fp1bis



"""
Test module for bundle_loader.py
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os.path as pth

import pelix
import pytest
from pelix.framework import FrameworkFactory

from .. import BundleLoader
from ..exceptions import FastDuplicateFactoryError

_LOGGER = logging.getLogger(__name__)
"""Logger for this module"""

logging.basicConfig(level=logging.DEBUG)


@pytest.fixture()
def delete_framework():
    """ Ensures framework is deleted before and after running tests"""
    if FrameworkFactory.is_framework_running():
        FrameworkFactory.get_framework().delete(True)

    yield

    if FrameworkFactory.is_framework_running():
        FrameworkFactory.get_framework().delete(True)


def test_init_bundle_loader_from_scratch(delete_framework):
    """
    Tests that creating a Loader instance will start a correctly configured
    Pelix framework
    """
    assert not FrameworkFactory.is_framework_running()

    loader = BundleLoader()
    _ = loader.framework  # a first call for initiating the framework
    framework = FrameworkFactory.get_framework()
    assert FrameworkFactory.is_framework_running()
    assert framework is loader.framework
    assert framework.get_bundle_by_name("pelix.ipopo.core") is not None

    # Framework outlives the Loader
    del loader
    assert FrameworkFactory.is_framework_running()

    # Another instance gets back the framework
    loader = BundleLoader()
    assert framework is loader.framework


def test_init_bundle_loader_with_running_framework(delete_framework):
    """
    Tests that creating a Loader instance with an already running Pelix
    framework will work and enforce a correct configuration of the framework.
    """

    framework = pelix.framework.create_framework(())
    assert framework.get_bundle_by_name("pelix.ipopo.core") is None
    assert FrameworkFactory.is_framework_running()

    # Creating a loader instance where existing framework misses
    # pelix.ipopo.core bundle.
    # This bundle is expected to be added to the framework at Loader
    # initialization
    loader = BundleLoader()
    assert framework is loader.framework
    assert loader.framework.get_bundle_by_name("pelix.ipopo.core") is not None
    del loader

    # Same as above, but now, existing framework HAS pelix.ipopo.core bundle
    loader = BundleLoader()
    assert framework is loader.framework
    assert loader.framework.get_bundle_by_name("pelix.ipopo.core") is not None
    del loader

    # Framework outlives the Loader
    assert FrameworkFactory.is_framework_running()


def test_install_packages(delete_framework):
    """
    Tests installation of bundles in a folder
    """
    loader = BundleLoader()

    loader.install_packages(pth.join(pth.dirname(__file__), "dummy_pelix_bundles"))
    assert (
        loader.framework.get_bundle_by_name("dummy_pelix_bundles.hello_world_with_decorators")
        is not None
    )


def test_install_packages_on_faulty_install(delete_framework):
    """
    Related to Issue #81
    """
    # TODO: this test was done for testing the workaround about iPOPO crashing
    #       when a module has None as __path__ attribute. As iPOPO is fixed, the
    #       workaround has been removed, and this test now just tests iPOPO, so
    #       it may be removed later.
    loader = BundleLoader()

    # Create the buggy numpy install
    import sys

    sys.modules["numpy.random.mtrand"].__path__ = None

    # Install packages
    loader.install_packages(pth.join(pth.dirname(__file__), "dummy_pelix_bundles"))
    assert (
        loader.framework.get_bundle_by_name("dummy_pelix_bundles.hello_world_with_decorators")
        is not None
    )


def test_register_factory(delete_framework):
    """
    Tests that register_factory raises correctly an error when finding
    duplicate names
    (Real tests are done when trying to get components registered in
    hello_world_without_decorators.py)
    """

    loader = BundleLoader()
    loader.install_packages(pth.join(pth.dirname(__file__), "dummy_pelix_bundles"))

    class Greetings1:
        def hello(self, name="World"):
            return "Hello, {0}!".format(name)

    with pytest.raises(FastDuplicateFactoryError) as exc_info:
        loader.register_factory(Greetings1, "hello-universe-factory", "hello.world")
    assert exc_info.value.factory_name == "hello-universe-factory"


def test_get_services(delete_framework):
    """
    Tests the method for retrieving services according to properties
    """
    loader = BundleLoader()
    loader.install_packages(pth.join(pth.dirname(__file__), "dummy_pelix_bundles"))

    # Missing service
    services = loader.get_services("does.not.exists")
    assert services is None

    # Existing service, but not instantiated
    services = loader.get_services("hello.world.no.instance")
    assert services is None

    # Existing services, no property provided
    services = loader.get_services("hello.world")
    assert len(services) == 2

    # Existing services, 1 property provided
    services = loader.get_services("hello.world", {"Instantiated": True})
    assert len(services) == 2
    services = loader.get_services("hello.world", {"Prop1": 1})
    assert len(services) == 1

    # Existing service, 2 properties, case insensitivity
    services = loader.get_services("hello.world", {"Prop1": 1, "Prop 2": "says.hello"})
    assert len(services) == 1
    greet_service = services[0]
    assert greet_service.hello("Dolly") == "Hello, Dolly!"

    # Existing service, case insensitivity
    services = loader.get_services("hello.world", {"Prop 2": "SAYS.HI"})
    assert len(services) == 1
    greet_service = services[0]
    assert greet_service.hello() == "Hi, World!"

    # Existing service, case sensitivity
    services = loader.get_services("hello.world", {"Prop 2": "SAYS.HELLO"}, True)
    assert services is None
    services = loader.get_services("hello.world", {"Prop 2": "Says.Hello"}, True)
    assert len(services) == 1
    river = services[0]
    assert river.hello("Sweetie") == "Hello, Sweetie!"


def test_instantiate_component(delete_framework):
    """
    Tests the method for instantiating a component from factory name
    """
    loader = BundleLoader()
    loader.install_packages(pth.join(pth.dirname(__file__), "dummy_pelix_bundles"))

    # 2 services should already be instantiated
    services = loader.get_services("hello.world")
    assert len(services) == 2

    song = loader.instantiate_component("another-hello-world-factory")
    assert song.hello("Sweetie") == "Hello again, Sweetie!"
    assert song._Prop1 == 3
    assert song._Prop_2 == "Says.Hello"

    # now one more service is instantiated
    services = loader.get_services("hello.world")
    assert len(services) == 3


def test_get_factory_names(delete_framework):
    """
    Tests the method for retrieving factories according to properties
    """
    loader = BundleLoader()
    loader.install_packages(pth.join(pth.dirname(__file__), "dummy_pelix_bundles"))

    # Missing service
    factory_names = loader.get_factory_names("does.not.exists")
    assert not factory_names

    # Get all factories
    factory_names = loader.get_factory_names()
    assert len(factory_names) >= 4  # the four test factories + factories in FAST

    # Existing services, no property provided
    factory_names = loader.get_factory_names("hello.world")
    assert len(factory_names) == 3
    factory_names = loader.get_factory_names("hello.universe")
    assert len(factory_names) == 1

    # No service provided, but properties
    factory_names = loader.get_factory_names(properties={"Instantiated": False})
    assert len(factory_names) == 1

    # Existing services, property provided does not apply
    factory_names = loader.get_factory_names("hello.world", {"Not A Property": 1})
    assert len(factory_names) == 0

    # Existing services, 1 property provided
    factory_names = loader.get_factory_names("hello.world", {"Instantiated": True})
    assert len(factory_names) == 2
    factory_names = loader.get_factory_names("hello.world", {"Prop1": 1})
    assert len(factory_names) == 1

    # Existing service, 2 properties, case insensitivity
    factory_names = loader.get_factory_names("hello.world", {"Prop1": 1, "Prop 2": "says.hello"})
    assert len(factory_names) == 1
    greet_service = loader.instantiate_component(factory_names[0])
    assert greet_service.hello("Dolly") == "Hello, Dolly!"

    # Existing service, case insensitivity
    factory_names = loader.get_factory_names("hello.world", {"Prop 2": "SAYS.HI"})
    assert len(factory_names) == 1
    greet_service = loader.instantiate_component(factory_names[0])
    assert greet_service.hello() == "Hi, World!"

    # Existing service, case sensitivity
    factory_names = loader.get_factory_names("hello.world", {"Prop 2": "SAYS.HELLO"}, True)
    assert not factory_names
    factory_names = loader.get_factory_names("hello.world", {"Prop 2": "Says.Hello"}, True)
    assert len(factory_names) == 2



"""
    FAST - Copyright (c) 2016 ONERA ISAE
"""

#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np
from fastoad.utils.physics import AtmosphereSI
from openmdao.core.explicitcomponent import ExplicitComponent


class ComputeReynolds(ExplicitComponent):
    def initialize(self):
        self.options.declare("low_speed_aero", default=False, types=bool)

    def setup(self):
        self.low_speed_aero = self.options["low_speed_aero"]

        if self.low_speed_aero:
            self.add_input("data:aerodynamics:aircraft:takeoff:mach", val=np.nan)
            self.add_output("data:aerodynamics:wing:low_speed:reynolds")
        else:
            self.add_input("data:TLAR:cruise_mach", val=np.nan)
            self.add_input("data:mission:sizing:main_route:cruise:altitude", val=np.nan, units="m")
            self.add_output("data:aerodynamics:wing:cruise:reynolds")

        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs):
        if self.low_speed_aero:
            mach = inputs["data:aerodynamics:aircraft:takeoff:mach"]
            altitude = 0.0
        else:
            mach = inputs["data:TLAR:cruise_mach"]
            altitude = inputs["data:mission:sizing:main_route:cruise:altitude"]

        reynolds = AtmosphereSI(altitude).get_unitary_reynolds(mach)

        if self.low_speed_aero:
            outputs["data:aerodynamics:wing:low_speed:reynolds"] = reynolds
        else:
            outputs["data:aerodynamics:wing:cruise:reynolds"] = reynolds



#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_allclose
from scipy.constants import foot

from fastoad.base.flight_point import FlightPoint
from fastoad.models.performances.mission.segments.hold import HoldSegment
from fastoad.models.propulsion.fuel_propulsion.base import AbstractFuelPropulsion, FuelEngineSet
from ..altitude_change import AltitudeChangeSegment
from ..cruise import OptimalCruiseSegment, CruiseSegment
from ..speed_change import SpeedChangeSegment
from ..taxi import TaxiSegment
from ...polar import Polar


def print_dataframe(df):
    """Utility for correctly printing results"""
    # Not used if all goes all well. Please keep it for future test setting/debugging.
    with pd.option_context(
        "display.max_rows", 20, "display.max_columns", None, "display.width", None
    ):
        print()
        print(df)


class DummyEngine(AbstractFuelPropulsion):
    def __init__(self, max_thrust, max_sfc):
        """
        Dummy engine model.

        Max thrust does not depend on flight conditions.
        SFC varies linearly with thrust_rate, from max_sfc/2. at thrust rate is 0.,
        to max_sfc when thrust_rate is 1.0

        :param max_thrust: thrust when thrust rate = 1.0
        :param max_sfc: SFC when thrust rate = 1.0
        """
        self.max_thrust = max_thrust
        self.max_sfc = max_sfc

    def compute_flight_points(self, flight_point: FlightPoint):

        if flight_point.thrust_is_regulated or flight_point.thrust_rate is None:
            flight_point.thrust_rate = flight_point.thrust / self.max_thrust
        else:
            flight_point.thrust = self.max_thrust * flight_point.thrust_rate

        flight_point.sfc = self.max_sfc * (1.0 + flight_point.thrust_rate) / 2.0


@pytest.fixture
def polar() -> Polar:
    """Returns a dummy polar where max L/D ratio is around 16."""
    cl = np.arange(0.0, 1.5, 0.01)
    cd = 0.5e-1 * cl ** 2 + 0.01
    return Polar(cl, cd)


def test_climb_fixed_altitude_at_constant_TAS(polar):
    propulsion = FuelEngineSet(DummyEngine(1.0e5, 1.0e-5), 2)

    # initialisation then change instance attributes
    segment = AltitudeChangeSegment(
        target={"altitude": 10000.0}, propulsion=propulsion, reference_area=120.0, polar=polar,
    )  # not constant TAS order, as it is the default
    segment.thrust_rate = 1.0
    segment.time_step = 2.0
    flight_points = segment.compute_from(
        {"altitude": 5000.0, "mass": 70000.0, "true_airspeed": 150.0}
    )  # Test with dict

    last_point = flight_points.iloc[-1]
    # Note: reference values are obtained by running the process with 0.01s as time step
    assert_allclose(last_point.altitude, 10000.0)
    assert_allclose(last_point.true_airspeed, 150.0)
    assert_allclose(last_point.time, 143.5, rtol=1e-2)
    assert_allclose(last_point.mass, 69713.0, rtol=1e-4)
    assert_allclose(last_point.ground_distance, 20943.0, rtol=1e-3)


def test_climb_fixed_altitude_at_constant_EAS(polar):
    propulsion = FuelEngineSet(DummyEngine(1.0e5, 1.0e-5), 2)

    flight_points = AltitudeChangeSegment(
        target=FlightPoint(altitude=10000.0, equivalent_airspeed="constant"),
        propulsion=propulsion,
        reference_area=120.0,
        polar=polar,
        thrust_rate=1.0,
        time_step=2.0,
    ).compute_from(FlightPoint(altitude=5000.0, mass=70000.0, equivalent_airspeed=100.0))

    first_point = flight_points.iloc[0]
    last_point = flight_points.iloc[-1]
    # Note: reference values are obtained by running the process with 0.01s as time step
    assert_allclose(last_point.altitude, 10000.0)
    assert_allclose(last_point.equivalent_airspeed, 100.0)
    assert_allclose(last_point.time, 145.2, rtol=1e-2)
    assert_allclose(first_point.true_airspeed, 129.0, atol=0.1)
    assert_allclose(last_point.true_airspeed, 172.3, atol=0.1)
    assert_allclose(last_point.mass, 69710.0, rtol=1e-4)
    assert_allclose(last_point.ground_distance, 20915.0, rtol=1e-3)


def test_climb_optimal_altitude_at_fixed_TAS(polar):
    propulsion = FuelEngineSet(DummyEngine(1.0e5, 1.0e-5), 2)

    flight_points = AltitudeChangeSegment(
        target=FlightPoint(
            altitude=AltitudeChangeSegment.OPTIMAL_ALTITUDE, true_airspeed="constant"
        ),
        propulsion=propulsion,
        reference_area=120.0,
        polar=polar,
        thrust_rate=1.0,
        time_step=2.0,
    ).compute_from(FlightPoint(altitude=5000.0, true_airspeed=250.0, mass=70000.0),)

    last_point = flight_points.iloc[-1]
    # Note: reference values are obtained by running the process with 0.01s as time step
    assert_allclose(last_point.altitude, 10085.0, atol=0.1)
    assert_allclose(last_point.true_airspeed, 250.0)
    assert_allclose(last_point.time, 84.1, rtol=1e-2)
    assert_allclose(last_point.mach, 0.8359, rtol=1e-4)
    assert_allclose(last_point.mass, 69832.0, rtol=1e-4)
    assert_allclose(last_point.ground_distance, 20401.0, rtol=1e-3)


def test_climb_optimal_flight_level_at_fixed_TAS(polar):
    propulsion = FuelEngineSet(DummyEngine(1.0e5, 1.0e-5), 2)

    flight_points = AltitudeChangeSegment(
        target=FlightPoint(
            altitude=AltitudeChangeSegment.OPTIMAL_FLIGHT_LEVEL, true_airspeed="constant"
        ),
        propulsion=propulsion,
        reference_area=120.0,
        polar=polar,
        thrust_rate=1.0,
        time_step=2.0,
    ).compute_from(FlightPoint(altitude=5000.0, true_airspeed=250.0, mass=70000.0),)
    print_dataframe(flight_points)

    last_point = flight_points.iloc[-1]
    # Note: reference values are obtained by running the process with 0.01s as time step
    assert_allclose(last_point.altitude / foot, 32000.0, atol=0.1)
    assert_allclose(last_point.true_airspeed, 250.0)
    assert_allclose(last_point.time, 78.7, rtol=1e-2)
    assert_allclose(last_point.mach, 0.8318, rtol=1e-4)
    assert_allclose(last_point.mass, 69843.0, rtol=1e-4)
    assert_allclose(last_point.ground_distance, 19091.0, rtol=1e-3)


def test_climb_optimal_altitude_at_fixed_TAS_with_capped_mach(polar):
    propulsion = FuelEngineSet(DummyEngine(1.0e5, 1.0e-5), 2)

    flight_points = AltitudeChangeSegment(
        target=FlightPoint(
            altitude=AltitudeChangeSegment.OPTIMAL_ALTITUDE, true_airspeed="constant"
        ),
        propulsion=propulsion,
        reference_area=120.0,
        polar=polar,
        thrust_rate=1.0,
        time_step=2.0,
        maximum_mach=0.80,
    ).compute_from(FlightPoint(altitude=5000.0, true_airspeed=250.0, mass=70000.0),)

    last_point = flight_points.iloc[-1]
    # Note: reference values are obtained by running the process with 0.01s as time step
    assert_allclose(last_point.CL, polar.optimal_cl)
    assert_allclose(last_point.altitude, 9507.2, atol=0.1)
    assert_allclose(last_point.mach, 0.80, rtol=1e-4)
    assert_allclose(last_point.time, 75.4, rtol=1e-2)
    assert_allclose(last_point.true_airspeed, 241.3, atol=0.1)
    assert_allclose(last_point.mass, 69849.0, rtol=1e-4)
    assert_allclose(last_point.ground_distance, 18112.0, rtol=1e-3)


def test_climb_not_enough_thrust(polar):
    propulsion = FuelEngineSet(DummyEngine(1.0e5, 1.0e-5), 2)

    segment = AltitudeChangeSegment(
        target=FlightPoint(altitude=10000.0, true_airspeed="constant"),
        propulsion=propulsion,
        reference_area=120.0,
        polar=polar,
        thrust_rate=0.1,
    )
    assert (
        len(segment.compute_from(FlightPoint(altitude=5000.0, true_airspeed=150.0, mass=70000.0),))
        == 1
    )


def test_descent_to_fixed_altitude_at_constant_TAS(polar):
    propulsion = FuelEngineSet(DummyEngine(1.0e5, 1.0e-5), 2)

    flight_points = AltitudeChangeSegment(
        target=FlightPoint(altitude=5000.0, true_airspeed="constant"),
        propulsion=propulsion,
        reference_area=100.0,
        polar=polar,
        thrust_rate=0.1,
        time_step=2.0,
    ).compute_from(
        FlightPoint(altitude=10000.0, true_airspeed=200.0, mass=70000.0, time=2000.0),
    )  # And we define a non-null start time

    last_point = flight_points.iloc[-1]
    # Note: reference values are obtained by running the process with 0.01s as time step
    assert_allclose(last_point.altitude, 5000.0)
    assert_allclose(last_point.true_airspeed, 200.0)
    assert_allclose(last_point.time, 3370.4, rtol=1e-2)
    assert_allclose(last_point.mass, 69849.0, rtol=1e-4)
    assert_allclose(last_point.ground_distance, 274043.0, rtol=1e-3)


def test_descent_to_fixed_altitude_at_constant_EAS(polar):
    propulsion = FuelEngineSet(DummyEngine(1.0e5, 1.0e-5), 2)

    flight_points = AltitudeChangeSegment(
        target=FlightPoint(altitude=5000.0, equivalent_airspeed="constant"),
        propulsion=propulsion,
        reference_area=100.0,
        polar=polar,
        thrust_rate=0.1,
        time_step=2.0,
    ).compute_from(FlightPoint(altitude=10000.0, equivalent_airspeed=200.0, mass=70000.0),)

    last_point = flight_points.iloc[-1]
    # Note: reference values are obtained by running the process with 0.01s as time step
    assert_allclose(last_point.altitude, 5000.0)
    assert_allclose(last_point.equivalent_airspeed, 200.0)
    assert_allclose(last_point.time, 821.4, rtol=1e-2)
    assert_allclose(last_point.mass, 69910.0, rtol=1e-4)
    assert_allclose(last_point.ground_distance, 243155.0, rtol=1e-3)


def test_descent_to_fixed_EAS_at_constant_mach(polar):
    propulsion = FuelEngineSet(DummyEngine(1.0e5, 1.0e-5), 2)

    flight_points = AltitudeChangeSegment(
        target=FlightPoint(equivalent_airspeed=150.0, mach="constant"),
        propulsion=propulsion,
        reference_area=100.0,
        polar=polar,
        thrust_rate=0.1,
        # time_step=5.0, # we use default time step
    ).compute_from(FlightPoint(altitude=10000.0, mass=70000.0, mach=0.78),)

    last_point = flight_points.iloc[-1]
    # Note: reference values are obtained by running the process with 0.01s as time step
    assert_allclose(last_point.equivalent_airspeed, 150.0)
    assert_allclose(last_point.mach, 0.78)
    assert_allclose(last_point.time, 343.6, rtol=1e-2)
    assert_allclose(last_point.altitude, 8654.0, atol=1.0)
    assert_allclose(last_point.true_airspeed, 238.1, atol=0.1)
    assert_allclose(last_point.mass, 69962.0, rtol=1e-4)
    assert_allclose(last_point.ground_distance, 81042.0, rtol=1e-3)


def test_acceleration_to_TAS(polar):
    propulsion = FuelEngineSet(DummyEngine(0.5e5, 1.0e-5), 2)

    segment = SpeedChangeSegment(
        target={"true_airspeed": 250.0},
        propulsion=propulsion,
        reference_area=120.0,
        polar=polar,
        thrust_rate=1.0,
        time_step=0.2,
    )
    flight_points = segment.compute_from(
        {"altitude": 5000.0, "true_airspeed": 150.0, "mass": 70000.0}
    )  # Test with dict

    last_point = flight_points.iloc[-1]
    # Note: reference values are obtained by running the process with 0.01s as time step
    assert_allclose(last_point.time, 103.3, rtol=1e-2)
    assert_allclose(last_point.altitude, 5000.0)
    assert_allclose(last_point.true_airspeed, 250.0)
    assert_allclose(last_point.mass, 69896.0, rtol=1e-4)
    assert_allclose(last_point.ground_distance, 20697.0, rtol=1e-3)


def test_acceleration_to_EAS(polar):
    propulsion = FuelEngineSet(DummyEngine(0.5e5, 1.0e-5), 2)

    flight_points = SpeedChangeSegment(
        target=FlightPoint(equivalent_airspeed=250.0),
        propulsion=propulsion,
        reference_area=120.0,
        polar=polar,
        thrust_rate=1.0,
        time_step=0.2,
    ).compute_from(FlightPoint(altitude=1000.0, true_airspeed=150.0, mass=70000.0))

    last_point = flight_points.iloc[-1]
    # Note: reference values are obtained by running the process with 0.01s as time step
    assert_allclose(last_point.time, 128.2, rtol=1e-2)
    assert_allclose(last_point.altitude, 1000.0)
    assert_allclose(last_point.true_airspeed, 262.4, atol=1e-1)
    assert_allclose(last_point.mass, 69872.0, rtol=1e-4)
    assert_allclose(last_point.ground_distance, 26868.0, rtol=1e-3)


def test_acceleration_to_mach(polar):
    propulsion = FuelEngineSet(DummyEngine(0.5e5, 1.0e-5), 2)

    flight_points = SpeedChangeSegment(
        target=FlightPoint(mach=0.8),
        propulsion=propulsion,
        reference_area=120.0,
        polar=polar,
        thrust_rate=1.0,
        time_step=0.2,
    ).compute_from(FlightPoint(altitude=1000.0, true_airspeed=150.0, mass=70000.0))

    last_point = flight_points.iloc[-1]
    # Note: reference values are obtained by running the process with 0.01s as time step
    assert_allclose(last_point.time, 138.0, rtol=1e-2)
    assert_allclose(last_point.altitude, 1000.0)
    assert_allclose(last_point.mach, 0.8, atol=1e-5)
    assert_allclose(last_point.mass, 69862.0, rtol=1e-4)
    assert_allclose(last_point.ground_distance, 29470.0, rtol=1e-3)


def test_acceleration_not_enough_thrust(polar):
    propulsion = FuelEngineSet(DummyEngine(0.5e5, 1.0e-5), 2)

    segment = SpeedChangeSegment(
        target=FlightPoint(true_airspeed=250.0),
        propulsion=propulsion,
        reference_area=120.0,
        polar=polar,
        thrust_rate=0.1,
    )
    assert len(
        segment.compute_from(FlightPoint(altitude=5000.0, true_airspeed=150.0, mass=70000.0),)
    )


def test_deceleration_not_enough_thrust(polar):
    propulsion = FuelEngineSet(DummyEngine(0.5e5, 1.0e-5), 2)

    segment = SpeedChangeSegment(
        target=FlightPoint(true_airspeed=150.0),
        propulsion=propulsion,
        reference_area=120.0,
        polar=polar,
        thrust_rate=0.1,
        time_step=1.0,
    )
    segment.time_step = 1.0
    flight_points = segment.compute_from(
        FlightPoint(altitude=5000.0, true_airspeed=250.0, mass=70000.0),
    )

    last_point = flight_points.iloc[-1]
    # Note: reference values are obtained by running the process with 0.01s as time step
    assert_allclose(last_point.time, 315.8, rtol=1e-2)
    assert_allclose(last_point.altitude, 5000.0)
    assert_allclose(last_point.true_airspeed, 150.0)
    assert_allclose(last_point.mass, 69982.0, rtol=1e-4)
    assert_allclose(last_point.ground_distance, 62804.0, rtol=1e-3)


def test_cruise_at_constant_altitude(polar):
    propulsion = FuelEngineSet(DummyEngine(0.5e5, 1.0e-5), 2)

    segment = CruiseSegment(
        target=FlightPoint(ground_distance=5.0e5),
        propulsion=propulsion,
        reference_area=120.0,
        polar=polar,
    )
    flight_points = segment.compute_from(FlightPoint(mass=70000.0, altitude=10000.0, mach=0.78))

    print_dataframe(flight_points)

    first_point = flight_points.iloc[0]
    last_point = FlightPoint(flight_points.iloc[-1])
    # Note: reference values are obtained by running the process with 0.05s as time step
    assert_allclose(first_point.altitude, 10000.0)
    assert_allclose(first_point.true_airspeed, 233.6, atol=0.1)

    assert_allclose(last_point.ground_distance, 500000.0)
    assert_allclose(last_point.altitude, 10000.0)
    assert_allclose(last_point.time, 2141.0, rtol=1e-2)
    assert_allclose(last_point.true_airspeed, 233.6, atol=0.1)
    assert_allclose(last_point.mass, 69568.0, rtol=1e-4)


def test_optimal_cruise(polar):
    propulsion = FuelEngineSet(DummyEngine(0.5e5, 1.0e-5), 2)

    segment = OptimalCruiseSegment(
        target=FlightPoint(ground_distance=5.0e5),
        propulsion=propulsion,
        reference_area=120.0,
        polar=polar,
    )
    flight_points = segment.compute_from(
        FlightPoint(mass=70000.0, time=1000.0, ground_distance=1e5, mach=0.78),
    )

    first_point = flight_points.iloc[0]
    last_point = FlightPoint(flight_points.iloc[-1])
    # Note: reference values are obtained by running the process with 0.05s as time step
    assert_allclose(first_point.altitude, 9156.0, atol=1.0)
    assert_allclose(first_point.true_airspeed, 236.4, atol=0.1)
    assert_allclose(first_point.CL, polar.optimal_cl)

    assert_allclose(last_point.ground_distance, 600000.0)
    assert_allclose(last_point.CL, polar.optimal_cl)
    assert_allclose(last_point.altitude, 9196.0, atol=1.0)
    assert_allclose(last_point.time, 3115.0, rtol=1e-2)
    assert_allclose(last_point.true_airspeed, 236.3, atol=0.1)
    assert_allclose(last_point.mass, 69577.0, rtol=1e-4)


def test_taxi():
    propulsion = FuelEngineSet(DummyEngine(0.5e5, 1.0e-5), 2)

    segment = TaxiSegment(target=FlightPoint(time=500.0), propulsion=propulsion, thrust_rate=0.1)
    flight_points = segment.compute_from(
        FlightPoint(altitude=10.0, true_airspeed=10.0, mass=50000.0, time=10000.0),
    )

    last_point = FlightPoint(flight_points.iloc[-1])
    assert_allclose(last_point.altitude, 10.0, atol=1.0)
    assert_allclose(last_point.time, 10500.0, rtol=1e-2)
    assert_allclose(last_point.true_airspeed, 10.0, atol=0.1)
    assert_allclose(last_point.mass, 49973.0, rtol=1e-4)
    assert_allclose(last_point.ground_distance, 5000.0)


def test_hold(polar):
    propulsion = FuelEngineSet(DummyEngine(0.5e5, 2.0e-5), 2)

    segment = HoldSegment(
        target=FlightPoint(time=3000.0), propulsion=propulsion, reference_area=120.0, polar=polar
    )
    flight_points = segment.compute_from(
        FlightPoint(altitude=500.0, equivalent_airspeed=250.0, mass=60000.0),
    )

    print_dataframe(flight_points)

    last_point = FlightPoint(flight_points.iloc[-1])
    assert_allclose(last_point.time, 3000.0)
    assert_allclose(last_point.altitude, 500.0)
    assert_allclose(last_point.equivalent_airspeed, 250.0, atol=0.1)
    assert_allclose(last_point.mass, 57776.0, rtol=1e-4)
    assert_allclose(last_point.ground_distance, 768323.0, rtol=1.0e-3)



#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os.path as pth
from logging import FileHandler
from os import mkdir
from shutil import rmtree

import openmdao.api as om
import pytest
from fastoad.openmdao.validity_checker import ValidityDomainChecker, ValidityStatus
from fastoad.openmdao.variables import VariableList, Variable

_LOGGER = logging.getLogger(__name__)  # Logger for this module

DATA_FOLDER_PATH = pth.join(pth.dirname(__file__), "data")
RESULTS_FOLDER_PATH = pth.join(pth.dirname(__file__), "results")


@pytest.fixture(scope="module")
def cleanup():
    rmtree(RESULTS_FOLDER_PATH, ignore_errors=True)
    mkdir(RESULTS_FOLDER_PATH)


def set_logger_file(log_file_path):
    logger = logging.getLogger("main")
    for hdlr in logger.handlers:
        logger.removeHandler(hdlr)

    formatter = logging.Formatter("%(name)s - %(message)s")
    handler = FileHandler(log_file_path)
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def test_register_checks_instantiation(cleanup):
    ValidityDomainChecker(
        {
            "var_with_upper_and_lower": (-1.0, 10.0),
            "var_with_lower": (-10.0, None),
            "var_with_upper": (None, 5.0),
        },
        "main.logger1",
    )

    # Now registering is done through instantiation, and with another logger
    ValidityDomainChecker(
        {"var_with_upper_and_lower": (-3.0, 5.0), "var_with_lower": (0.0, None)}, "main.logger2"
    )

    # Third call with same logger as the first one
    ValidityDomainChecker({"other_var": (-10.0, 1.0)}, "main.logger1")

    # Check 1 --------------------------------------------------------------------------------------
    log_file_path = pth.join(RESULTS_FOLDER_PATH, "log1.txt")
    set_logger_file(log_file_path)
    variables = VariableList(
        [
            Variable("var_with_upper_and_lower", value=-2.0, units="m"),  # too low for logger1
            Variable("var_with_lower", value=-1.0),  # too low for logger2
            Variable("var_with_upper", value=10.0),  # too high, only for logger1
            Variable("other_var", value=1.1),  # too high, only for logger1 (second registering)
            Variable("unbound_var", value=42.0),
        ]
    )

    records = ValidityDomainChecker.check_variables(variables)
    assert [
        (
            rec.variable_name,
            rec.status,
            rec.limit_value,
            rec.value,
            rec.source_file,
            rec.logger_name,
        )
        for rec in records
    ] == [
        ("var_with_upper_and_lower", ValidityStatus.TOO_LOW, -1.0, -2.0, __file__, "main.logger1"),
        ("var_with_upper_and_lower", ValidityStatus.OK, None, -2.0, __file__, "main.logger2"),
        ("var_with_lower", ValidityStatus.OK, None, -1.0, __file__, "main.logger1"),
        ("var_with_lower", ValidityStatus.TOO_LOW, 0.0, -1.0, __file__, "main.logger2"),
        ("var_with_upper", ValidityStatus.TOO_HIGH, 5.0, 10.0, __file__, "main.logger1"),
        ("other_var", ValidityStatus.TOO_HIGH, 1.0, 1.1, __file__, "main.logger1"),
    ]

    ValidityDomainChecker.log_records(records)
    with open(log_file_path) as log_file:
        assert len(log_file.readlines()) == 4

    # Check 2 --------------------------------------------------------------------------------------
    log_file_path = pth.join(RESULTS_FOLDER_PATH, "log2.txt")
    set_logger_file(log_file_path)
    variables = VariableList(
        [
            Variable("other_var", value=-11.0),  # too low, only for logger1 (second registering)
            Variable("var_with_lower", value=-15.0),  # too low for logger1 and logger2
            Variable("var_with_upper", value=0.0),  # Ok
            Variable("unbound_var", value=1e42),
            Variable("var_with_upper_and_lower", value=7.0),  # too high for logger2
        ]
    )

    records = ValidityDomainChecker.check_variables(variables)
    assert [
        (
            rec.variable_name,
            rec.status,
            rec.limit_value,
            rec.value,
            rec.source_file,
            rec.logger_name,
        )
        for rec in records
    ] == [
        ("other_var", ValidityStatus.TOO_LOW, -10.0, -11.0, __file__, "main.logger1"),
        ("var_with_lower", ValidityStatus.TOO_LOW, -10.0, -15.0, __file__, "main.logger1"),
        ("var_with_lower", ValidityStatus.TOO_LOW, 0.0, -15.0, __file__, "main.logger2"),
        ("var_with_upper", ValidityStatus.OK, None, 0.0, __file__, "main.logger1"),
        ("var_with_upper_and_lower", ValidityStatus.OK, None, 7.0, __file__, "main.logger1"),
        ("var_with_upper_and_lower", ValidityStatus.TOO_HIGH, 5.0, 7.0, __file__, "main.logger2"),
    ]

    ValidityDomainChecker.log_records(records)
    with open(log_file_path) as log_file:
        assert len(log_file.readlines()) == 4

    # Check 3 --------------------------------------------------------------------------------------
    log_file_path = pth.join(RESULTS_FOLDER_PATH, "log3.txt")
    set_logger_file(log_file_path)
    variables = VariableList(
        [
            Variable("var_with_upper_and_lower", value=1.0),  # Ok
            Variable("other_var", value=-5.0),  # Ok
            Variable("unbound_var", value=-1e42),
            Variable("var_with_lower", value=1.0),  # Ok
        ]
    )

    records = ValidityDomainChecker.check_variables(variables)
    assert [
        (
            rec.variable_name,
            rec.status,
            rec.limit_value,
            rec.value,
            rec.source_file,
            rec.logger_name,
        )
        for rec in records
    ] == [
        ("var_with_upper_and_lower", ValidityStatus.OK, None, 1.0, __file__, "main.logger1"),
        ("var_with_upper_and_lower", ValidityStatus.OK, None, 1.0, __file__, "main.logger2"),
        ("other_var", ValidityStatus.OK, None, -5.0, __file__, "main.logger1"),
        ("var_with_lower", ValidityStatus.OK, None, 1.0, __file__, "main.logger1"),
        ("var_with_lower", ValidityStatus.OK, None, 1.0, __file__, "main.logger2"),
    ]

    ValidityDomainChecker.log_records(records)
    with open(log_file_path) as log_file:
        assert len(log_file.readlines()) == 0


def test_register_checks_as_decorator(cleanup):
    log_file_path = pth.join(RESULTS_FOLDER_PATH, "log4.txt")
    set_logger_file(log_file_path)

    @ValidityDomainChecker({"input1": (1.0, 5.0), "output2": (300.0, 500.0)}, "main.dec")
    class Comp1(om.ExplicitComponent):
        def setup(self):
            self.add_input("input1", 0.5, units="km")
            self.add_output("output1", 150.0, units="km/h", upper=130.0)
            self.add_output("output2", 200.0, units="K", lower=0.0)
            self.add_output("output3", 1000.0, units="kg", lower=0.0, upper=5000.0)

    comp = Comp1()

    # Just checking there is no side effect. VariableList.from_system() uses
    # setup(), even if it is made to have no effect, and ValidityDomainChecker
    # modifies setup(), so is is worth checking.
    variables = VariableList.from_system(comp)
    assert len(variables) == 4

    # Now for the real test
    # We test that upper and lower bounds are retrieved from OpenMDAO component,
    # overwritten when required and that units are correctly taken into account.
    comp.setup()

    variables = VariableList(
        [
            Variable("input1", value=3.0, units="m"),
            Variable("output1", value=40.0, units="m/s"),
            Variable("output2", value=310.0, units="degC"),
            Variable("output3", value=6.0, units="t"),
        ]
    )
    records = ValidityDomainChecker.check_variables(variables)
    assert [
        (
            rec.variable_name,
            rec.status,
            rec.limit_value,
            rec.value,
            rec.source_file,
            rec.logger_name,
        )
        for rec in records
    ] == [
        ("input1", ValidityStatus.TOO_LOW, 1.0, 3.0, __file__, "main.dec"),
        ("output1", ValidityStatus.TOO_HIGH, 130.0, 40.0, __file__, "main.dec"),
        ("output2", ValidityStatus.TOO_HIGH, 500.0, 310.0, __file__, "main.dec"),
        ("output3", ValidityStatus.TOO_HIGH, 5000.0, 6.0, __file__, "main.dec"),
    ]

    ValidityDomainChecker.log_records(records)
    with open(log_file_path) as log_file:
        assert len(log_file.readlines()) == 4



"""
Tests for analysis and plots functions
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os.path as pth

from fastoad.utils.postprocessing.analysis_and_plots import (
    wing_geometry_plot,
    drag_polar_plot,
    mass_breakdown_sun_plot,
    aircraft_geometry_plot,
    mass_breakdown_bar_plot,
)

DATA_FOLDER_PATH = pth.join(pth.dirname(__file__), "data")


def test_wing_geometry_plot():
    """
    Basic tests for testing the plotting.
    """

    filename = pth.join(DATA_FOLDER_PATH, "problem_outputs.xml")

    # First plot
    # This is a rudimentary test as plot are difficult to verify
    # The test will fail if an error is raised by the following line
    fig = wing_geometry_plot(filename)

    # First plot with name
    # This is a rudimentary test as plot are difficult to verify
    # The test will fail if an error is raised by the following line
    fig = wing_geometry_plot(filename, name="First plot")

    # Adding a plot to the previous fig
    # This is a rudimentary test as plot are difficult to verify
    # The test will fail if an error is raised by the following line
    fig = wing_geometry_plot(filename, name="Second plot", fig=fig)


def test_aircraft_geometry_plot():
    """
    Basic tests for testing the plotting.
    """

    filename = pth.join(DATA_FOLDER_PATH, "problem_outputs.xml")

    # First plot
    # This is a rudimentary test as plot are difficult to verify
    # The test will fail if an error is raised by the following line
    fig = aircraft_geometry_plot(filename)

    # First plot with name
    # This is a rudimentary test as plot are difficult to verify
    # The test will fail if an error is raised by the following line
    fig = aircraft_geometry_plot(filename, name="First plot")

    # Adding a plot to the previous fig
    # This is a rudimentary test as plot are difficult to verify
    # The test will fail if an error is raised by the following line
    fig = aircraft_geometry_plot(filename, name="Second plot", fig=fig)


def test_mass_breakdown_bar_plot():
    """
    Basic tests for testing the plotting.
    """

    filename = pth.join(DATA_FOLDER_PATH, "problem_outputs.xml")

    # First plot
    # This is a rudimentary test as plot are difficult to verify
    # The test will fail if an error is raised by the following line
    fig = mass_breakdown_bar_plot(filename)

    # First plot with name
    # This is a rudimentary test as plot are difficult to verify
    # The test will fail if an error is raised by the following line
    fig = mass_breakdown_bar_plot(filename, name="First plot")

    # Adding a plot to the previous fig
    # This is a rudimentary test as plot are difficult to verify
    # The test will fail if an error is raised by the following line
    _ = mass_breakdown_bar_plot(filename, name="Second plot", fig=fig)


def test_drag_polar_plot():
    """
    Basic tests for testing the plotting.
    """

    filename = pth.join(DATA_FOLDER_PATH, "problem_outputs.xml")

    # First plot
    # This is a rudimentary test as plot are difficult to verify
    # The test will fail if an error is raised by the following line
    _ = drag_polar_plot(filename)


def test_mass_breakdown_sun_plot():
    """
    Basic tests for testing the plotting.
    """

    filename = pth.join(DATA_FOLDER_PATH, "problem_outputs.xml")

    # First plot
    # This is a rudimentary test as plot are difficult to verify
    # The test will fail if an error is raised by the following line
    _ = mass_breakdown_sun_plot(filename)



"""
    Estimation of vertical tail sweeps
"""

#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import math

import numpy as np
from openmdao.core.explicitcomponent import ExplicitComponent


# TODO: HT and VT components are similar --> factorize
class ComputeVTSweep(ExplicitComponent):
    # TODO: Document equations. Cite sources
    """ Vertical tail sweeps estimation """

    def setup(self):
        self.add_input("data:geometry:vertical_tail:span", val=np.nan, units="m")
        self.add_input("data:geometry:vertical_tail:root:chord", val=np.nan, units="m")
        self.add_input("data:geometry:vertical_tail:tip:chord", val=np.nan, units="m")
        self.add_input("data:geometry:vertical_tail:sweep_25", val=np.nan, units="deg")

        self.add_output("data:geometry:vertical_tail:sweep_0", units="deg")
        self.add_output("data:geometry:vertical_tail:sweep_100", units="deg")

        self.declare_partials("data:geometry:vertical_tail:sweep_0", "*", method="fd")
        self.declare_partials("data:geometry:vertical_tail:sweep_100", "*", method="fd")

    def compute(self, inputs, outputs):
        root_chord = inputs["data:geometry:vertical_tail:root:chord"]
        tip_chord = inputs["data:geometry:vertical_tail:tip:chord"]
        sweep_25_vt = inputs["data:geometry:vertical_tail:sweep_25"]
        b_v = inputs["data:geometry:vertical_tail:span"]

        sweep_0_vt = (
            (
                math.pi / 2
                - math.atan(
                    b_v
                    / (
                        0.25 * root_chord
                        - 0.25 * tip_chord
                        + b_v * math.tan(sweep_25_vt / 180.0 * math.pi)
                    )
                )
            )
            / math.pi
            * 180.0
        )
        sweep_100_vt = (
            (
                math.pi / 2
                - math.atan(
                    b_v
                    / (
                        b_v * math.tan(sweep_25_vt / 180.0 * math.pi)
                        - 0.75 * root_chord
                        + 0.75 * tip_chord
                    )
                )
            )
            / math.pi
            * 180.0
        )

        outputs["data:geometry:vertical_tail:sweep_0"] = sweep_0_vt
        outputs["data:geometry:vertical_tail:sweep_100"] = sweep_100_vt



"""Simple Breguet-based module for performances."""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from scipy.constants import g

from fastoad import BundleLoader
from fastoad.constants import EngineSetting
from fastoad.models.propulsion.fuel_propulsion.base import FuelEngineSet
from fastoad.utils.physics import AtmosphereSI
from . import Breguet


class OMBreguet(om.Group):
    """
    Estimation of fuel consumption through Breguet formula.

    It uses a rough estimate of climb and descent phases.

    MTOW (Max TakeOff Weight) being an input, the model computes the ZFW (Zero Fuel
    Weight) considering that all fuel but the reserve has been consumed during the
    mission.
    This model does not ensure consistency with OWE (Operating Empty Weight).
    """

    def initialize(self):
        self.options.declare("propulsion_id", default=None, types=str, allow_none=True)

    # TODO: in a more general case, this module will link the starting mass to
    #   the ending mass. Could we make the module more generic ?
    def setup(self):
        if self.options["propulsion_id"] is None:
            self.add_subsystem("propulsion_link", _BreguetPropulsion(), promotes=["*"])
            self.add_subsystem("distances", _Distances(), promotes=["*"])
            self.add_subsystem("cruise_mass_ratio", _CruiseMassRatio(), promotes=["*"])
            self.add_subsystem("fuel_weights", _FuelWeightFromMTOW(), promotes=["*"])
        else:
            self.add_subsystem(
                "propulsion",
                BreguetWithPropulsion(propulsion_id=self.options["propulsion_id"]),
                promotes=["*"],
            )
        self.add_subsystem("consumption", _Consumption(), promotes=["*"])


class BreguetWithPropulsion(om.ExplicitComponent):
    def initialize(self):
        self.options.declare("propulsion_id", default="", types=str)

    def setup(self):
        self._engine_wrapper = BundleLoader().instantiate_component(self.options["propulsion_id"])
        self._engine_wrapper.setup(self)

        self.add_input("data:mission:sizing:main_route:cruise:altitude", np.nan, units="m")
        self.add_input("data:TLAR:cruise_mach", np.nan)
        self.add_input("data:weight:aircraft:MTOW", np.nan, units="kg")
        self.add_input("data:aerodynamics:aircraft:cruise:L_D_max", np.nan)
        self.add_input("data:geometry:propulsion:engine:count", 2)
        self.add_input("settings:mission:sizing:breguet:climb:mass_ratio", 0.97)
        self.add_input("settings:mission:sizing:breguet:descent:mass_ratio", 0.98)
        self.add_input("settings:mission:sizing:breguet:reserve:mass_ratio", 0.06)
        self.add_input("data:TLAR:range", np.nan, units="m")
        self.add_input("settings:mission:sizing:breguet:climb_descent_distance", 500.0e3, units="m")

        self.add_output("data:mission:sizing:main_route:climb:distance", units="m", ref=1e3)
        self.add_output("data:mission:sizing:main_route:cruise:distance", units="m", ref=1e3)
        self.add_output("data:mission:sizing:main_route:descent:distance", units="m", ref=1e3)
        self.add_output("data:mission:sizing:ZFW", units="kg", ref=1e4)
        self.add_output("data:mission:sizing:fuel", units="kg", ref=1e4)
        self.add_output("data:mission:sizing:main_route:fuel", units="kg", ref=1e4)
        self.add_output("data:mission:sizing:main_route:climb:fuel", units="kg", ref=1e4)
        self.add_output("data:mission:sizing:main_route:cruise:fuel", units="kg", ref=1e4)
        self.add_output("data:mission:sizing:main_route:descent:fuel", units="kg", ref=1e4)
        self.add_output("data:mission:sizing:fuel_reserve", units="kg", ref=1e4)

        self.add_output("data:propulsion:SFC", units="kg/s/N", ref=1e-4)
        self.add_output("data:propulsion:thrust_rate", lower=0.0, upper=1.0)
        self.add_output("data:propulsion:thrust", units="N", ref=1e5)

        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        propulsion_model = FuelEngineSet(
            self._engine_wrapper.get_model(inputs), inputs["data:geometry:propulsion:engine:count"]
        )

        breguet = Breguet(
            propulsion_model,
            inputs["data:aerodynamics:aircraft:cruise:L_D_max"],
            inputs["data:TLAR:cruise_mach"],
            inputs["data:mission:sizing:main_route:cruise:altitude"],
            inputs["settings:mission:sizing:breguet:climb:mass_ratio"],
            inputs["settings:mission:sizing:breguet:descent:mass_ratio"],
            inputs["settings:mission:sizing:breguet:reserve:mass_ratio"],
            inputs["settings:mission:sizing:breguet:climb_descent_distance"],
        )
        breguet.compute(
            inputs["data:weight:aircraft:MTOW"], inputs["data:TLAR:range"],
        )

        outputs["data:propulsion:SFC"] = breguet.sfc
        outputs["data:propulsion:thrust_rate"] = breguet.thrust_rate
        outputs["data:propulsion:thrust"] = breguet.thrust
        outputs["data:mission:sizing:ZFW"] = breguet.zfw
        outputs["data:mission:sizing:fuel"] = breguet.mission_fuel
        outputs["data:mission:sizing:main_route:fuel"] = breguet.flight_fuel
        outputs["data:mission:sizing:main_route:climb:fuel"] = breguet.climb_fuel
        outputs["data:mission:sizing:main_route:cruise:fuel"] = breguet.cruise_fuel
        outputs["data:mission:sizing:main_route:descent:fuel"] = breguet.descent_fuel
        outputs["data:mission:sizing:main_route:climb:distance"] = breguet.climb_distance
        outputs["data:mission:sizing:main_route:cruise:distance"] = breguet.cruise_distance
        outputs["data:mission:sizing:main_route:descent:distance"] = breguet.descent_distance
        outputs["data:mission:sizing:fuel_reserve"] = breguet.reserve_fuel


class _Consumption(om.ExplicitComponent):
    """
    Adds a variable for consumption/passenger/km
    """

    def setup(self):
        self.add_input("data:mission:sizing:fuel", np.nan, units="kg")
        self.add_input("data:TLAR:range", np.nan, units="km")
        self.add_input("data:TLAR:NPAX", np.nan)

        self.add_output("data:mission:sizing:fuel:unitary", units="kg/km")
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        fuel = inputs["data:mission:sizing:fuel"]
        npax = inputs["data:TLAR:NPAX"]
        distance = inputs["data:TLAR:range"]

        outputs["data:mission:sizing:fuel:unitary"] = fuel / npax / distance


class _BreguetPropulsion(om.ExplicitComponent):
    """
    Link with engine computation.
    """

    def setup(self):
        self.add_input("data:mission:sizing:main_route:cruise:altitude", np.nan, units="m")
        self.add_input("data:TLAR:cruise_mach", np.nan)
        self.add_input("data:weight:aircraft:MTOW", np.nan, units="kg")
        self.add_input("data:aerodynamics:aircraft:cruise:L_D_max", np.nan)
        self.add_input("data:geometry:propulsion:engine:count", 2)
        self.add_input("settings:mission:sizing:breguet:climb:mass_ratio", 0.97)

        self.add_output("data:propulsion:engine_setting", EngineSetting.CRUISE)
        self.add_output("data:propulsion:use_thrust_rate", False)
        self.add_output("data:propulsion:required_thrust_rate", 0.0, lower=0.0, upper=1.0)
        self.add_output("data:propulsion:required_thrust", units="N", ref=1e5)
        self.add_output("data:propulsion:altitude", units="m", ref=1e4)
        self.add_output("data:propulsion:mach")

        self.declare_partials("data:propulsion:required_thrust_rate", "*", method="fd")
        self.declare_partials("data:propulsion:required_thrust", "*", method="fd")
        self.declare_partials(
            "data:propulsion:altitude",
            "data:mission:sizing:main_route:cruise:altitude",
            method="fd",
        )
        self.declare_partials("data:propulsion:mach", "data:TLAR:cruise_mach", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        engine_count = inputs["data:geometry:propulsion:engine:count"]
        ld_ratio = inputs["data:aerodynamics:aircraft:cruise:L_D_max"]
        mtow = inputs["data:weight:aircraft:MTOW"]
        initial_cruise_mass = mtow * inputs["settings:mission:sizing:breguet:climb:mass_ratio"]

        # Variables for propulsion
        outputs["data:propulsion:altitude"] = inputs[
            "data:mission:sizing:main_route:cruise:altitude"
        ]
        outputs["data:propulsion:mach"] = inputs["data:TLAR:cruise_mach"]

        outputs["data:propulsion:required_thrust"] = (
            initial_cruise_mass / ld_ratio * g / engine_count
        )


class _FuelWeightFromMTOW(om.ExplicitComponent):
    """
    Estimation of fuel consumption through Breguet formula with a rough estimate
    of climb and descent phases.
    """

    def setup(self):
        self.add_input("data:TLAR:cruise_mach", np.nan)
        self.add_input("data:TLAR:range", np.nan, units="m")
        self.add_input("data:aerodynamics:aircraft:cruise:L_D_max", np.nan)
        self.add_input("data:propulsion:SFC", np.nan, units="kg/N/s")
        self.add_input("data:weight:aircraft:MTOW", np.nan, units="kg")
        self.add_input("data:mission:sizing:main_route:cruise:mass_ratio", np.nan)
        self.add_input("data:mission:sizing:main_route:cruise:altitude", np.nan, units="m")
        self.add_input("settings:mission:sizing:breguet:climb:mass_ratio", 0.97)
        self.add_input("settings:mission:sizing:breguet:descent:mass_ratio", 0.98)
        self.add_input("settings:mission:sizing:breguet:reserve:mass_ratio", 0.06)

        self.add_output("data:mission:sizing:ZFW", units="kg", ref=1e4)
        self.add_output("data:mission:sizing:fuel", units="kg", ref=1e4)
        self.add_output("data:mission:sizing:main_route:fuel", units="kg", ref=1e4)
        self.add_output("data:mission:sizing:main_route:climb:fuel", units="kg", ref=1e4)
        self.add_output("data:mission:sizing:main_route:cruise:fuel", units="kg", ref=1e4)
        self.add_output("data:mission:sizing:main_route:descent:fuel", units="kg", ref=1e4)
        self.add_output("data:mission:sizing:fuel_reserve", units="kg", ref=1e4)

        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        mtow = inputs["data:weight:aircraft:MTOW"]
        cruise_mass_ratio = inputs["data:mission:sizing:main_route:cruise:mass_ratio"]
        climb_mass_ratio = inputs["settings:mission:sizing:breguet:climb:mass_ratio"]
        descent_mass_ratio = inputs["settings:mission:sizing:breguet:descent:mass_ratio"]
        reserve_mass_ratio = inputs["settings:mission:sizing:breguet:reserve:mass_ratio"]

        flight_mass_ratio = cruise_mass_ratio * climb_mass_ratio * descent_mass_ratio
        zfw = mtow * flight_mass_ratio / (1.0 + reserve_mass_ratio)
        mission_fuel = mtow - zfw

        outputs["data:mission:sizing:ZFW"] = zfw

        outputs["data:mission:sizing:fuel"] = mission_fuel
        outputs["data:mission:sizing:main_route:fuel"] = mtow * (1.0 - flight_mass_ratio)
        outputs["data:mission:sizing:main_route:climb:fuel"] = mtow * (1.0 - climb_mass_ratio)
        outputs["data:mission:sizing:main_route:cruise:fuel"] = (
            mtow * climb_mass_ratio * (1.0 - cruise_mass_ratio)
        )
        outputs["data:mission:sizing:main_route:descent:fuel"] = (
            mtow * climb_mass_ratio * cruise_mass_ratio * (1.0 - descent_mass_ratio)
        )
        outputs["data:mission:sizing:fuel_reserve"] = zfw * reserve_mass_ratio


class _Distances(om.ExplicitComponent):
    """
    Rough estimation of distances for each flight engine_setting.
    """

    def setup(self):
        self.add_input("data:TLAR:range", np.nan, units="m")
        self.add_input("settings:mission:sizing:breguet:climb_descent_distance", 500.0e3, units="m")

        self.add_output("data:mission:sizing:main_route:climb:distance", units="m", ref=1e3)
        self.add_output("data:mission:sizing:main_route:cruise:distance", units="m", ref=1e3)
        self.add_output("data:mission:sizing:main_route:descent:distance", units="m", ref=1e3)

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        flight_range = inputs["data:TLAR:range"]
        climb_descent_distance = inputs["settings:mission:sizing:breguet:climb_descent_distance"]

        outputs["data:mission:sizing:main_route:cruise:distance"] = (
            flight_range - climb_descent_distance
        )
        outputs["data:mission:sizing:main_route:climb:distance"] = climb_descent_distance / 2.0
        outputs["data:mission:sizing:main_route:descent:distance"] = climb_descent_distance / 2.0


class _CruiseMassRatio(om.ExplicitComponent):
    """
    Estimation of fuel consumption through Breguet formula for a given cruise distance.
    """

    def setup(self):
        self.add_input("data:aerodynamics:aircraft:cruise:L_D_max", np.nan)
        self.add_input("data:propulsion:SFC", np.nan, units="kg/N/s")
        self.add_input("data:TLAR:cruise_mach", np.nan)
        self.add_input("data:mission:sizing:main_route:cruise:altitude", np.nan, units="m")
        self.add_input("data:mission:sizing:main_route:cruise:distance", np.nan, units="m")

        # The lower bound helps a lot for convergence
        self.add_output("data:mission:sizing:main_route:cruise:mass_ratio", lower=0.5, upper=1.0)

        self.declare_partials("data:mission:sizing:main_route:cruise:mass_ratio", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        atmosphere = AtmosphereSI(inputs["data:mission:sizing:main_route:cruise:altitude"])
        cruise_speed = atmosphere.speed_of_sound * inputs["data:TLAR:cruise_mach"]

        cruise_distance = inputs["data:mission:sizing:main_route:cruise:distance"]
        ld_ratio = inputs["data:aerodynamics:aircraft:cruise:L_D_max"]
        sfc = inputs["data:propulsion:SFC"]

        range_factor = cruise_speed * ld_ratio / g / sfc
        # During first iterations, SFC will be incorrect and range_factor may be too low,
        # resulting in null or too small cruise_mass_ratio.
        # Forcing cruise_mass_ratio to a minimum of 0.3 avoids problems and should not
        # harm (no airplane loses 70% of its weight from fuel consumption)
        cruise_mass_ratio = np.maximum(0.3, 1.0 / np.exp(cruise_distance / range_factor))

        outputs["data:mission:sizing:main_route:cruise:mass_ratio"] = cruise_mass_ratio



#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import openmdao.api as om
from fastoad.base.flight_point import FlightPoint
from fastoad.constants import EngineSetting
from fastoad.models.propulsion.fuel_propulsion.rubber_engine import (
    OMRubberEngineComponent,
    RubberEngine,
)
from numpy.testing import assert_allclose
from scipy.constants import foot

from tests.testing_utilities import run_system
from ..openmdao import OMBreguet


def test_breguet():
    # test 1
    ivc = om.IndepVarComp()
    ivc.add_output("data:mission:sizing:main_route:cruise:altitude", 35000, units="ft")
    ivc.add_output("data:TLAR:cruise_mach", 0.78)
    ivc.add_output("data:TLAR:range", 500, units="NM")
    ivc.add_output("data:TLAR:NPAX", 150)
    ivc.add_output("data:aerodynamics:aircraft:cruise:L_D_max", 16.0)
    ivc.add_output("data:propulsion:SFC", 1e-5, units="kg/N/s")
    ivc.add_output("data:weight:aircraft:MTOW", 74000, units="kg")

    problem = run_system(OMBreguet(), ivc)

    assert_allclose(problem["data:mission:sizing:ZFW"], 65617.0, rtol=1e-3)
    assert_allclose(problem["data:mission:sizing:fuel"], 8382.0, rtol=1e-3)
    assert_allclose(problem["data:mission:sizing:fuel:unitary"], 0.0604, rtol=1e-3)
    assert_allclose(problem["data:mission:sizing:fuel:unitary"], 0.0604, rtol=1e-3)
    assert_allclose(
        problem.get_val("data:mission:sizing:main_route:climb:distance", units="km"), 250.0
    )
    assert_allclose(
        problem.get_val("data:mission:sizing:main_route:descent:distance", units="km"), 250.0
    )
    assert_allclose(
        problem.get_val("data:mission:sizing:main_route:cruise:distance", units="km"), 426.0
    )
    # test 2
    ivc = om.IndepVarComp()
    ivc.add_output("data:mission:sizing:main_route:cruise:altitude", 35000, units="ft")
    ivc.add_output("data:TLAR:cruise_mach", 0.78)
    ivc.add_output("data:TLAR:range", 1500, units="NM")
    ivc.add_output("data:TLAR:NPAX", 120)
    ivc.add_output("data:aerodynamics:aircraft:cruise:L_D_max", 16.0)
    ivc.add_output("data:propulsion:SFC", 1e-5, units="kg/N/s")
    ivc.add_output("data:weight:aircraft:MTOW", 74000, units="kg")

    problem = run_system(OMBreguet(), ivc)

    assert_allclose(problem["data:mission:sizing:ZFW"], 62473.0, rtol=1e-3)
    assert_allclose(problem["data:mission:sizing:fuel"], 11526.0, rtol=1e-3)
    assert_allclose(problem["data:mission:sizing:fuel:unitary"], 0.0346, rtol=1e-3)

    # Check consistency of other outputs
    assert_allclose(
        problem["data:mission:sizing:fuel"],
        problem["data:mission:sizing:main_route:fuel"]
        + problem["data:mission:sizing:fuel_reserve"],
        rtol=1e-3,
    )
    assert_allclose(
        problem["data:mission:sizing:main_route:fuel"],
        problem["data:mission:sizing:main_route:climb:fuel"]
        + problem["data:mission:sizing:main_route:cruise:fuel"]
        + problem["data:mission:sizing:main_route:descent:fuel"],
        rtol=1e-3,
    )


def test_breguet_with_rubber_engine():
    ivc = om.IndepVarComp()
    ivc.add_output("data:mission:sizing:main_route:cruise:altitude", 35000, units="ft")
    ivc.add_output("data:TLAR:cruise_mach", 0.78)
    ivc.add_output("data:TLAR:range", 500, units="NM")
    ivc.add_output("data:TLAR:NPAX", 150)
    ivc.add_output("data:aerodynamics:aircraft:cruise:L_D_max", 16.0)
    ivc.add_output("data:weight:aircraft:MTOW", 74000, units="kg")

    ivc.add_output("data:propulsion:rubber_engine:bypass_ratio", 5)
    ivc.add_output("data:propulsion:rubber_engine:maximum_mach", 0.95)
    ivc.add_output("data:propulsion:rubber_engine:design_altitude", 35000, units="ft")
    ivc.add_output("data:propulsion:MTO_thrust", 100000, units="N")
    ivc.add_output("data:propulsion:rubber_engine:overall_pressure_ratio", 30)
    ivc.add_output("data:propulsion:rubber_engine:turbine_inlet_temperature", 1500, units="K")

    # With rubber engine OM component
    group = om.Group()
    group.add_subsystem("breguet", OMBreguet(), promotes=["*"])
    group.add_subsystem("engine", OMRubberEngineComponent(), promotes=["*"])
    group.nonlinear_solver = om.NonlinearBlockGS()
    problem = run_system(group, ivc)

    assert_allclose(problem["data:mission:sizing:ZFW"], 65076.0, atol=1)
    assert_allclose(problem["data:mission:sizing:fuel"], 8924.0, atol=1)
    assert_allclose(problem["data:mission:sizing:fuel:unitary"], 0.0642, rtol=1e-3)

    # With direct call to rubber engine
    problem2 = run_system(OMBreguet(propulsion_id="fastoad.wrapper.propulsion.rubber_engine"), ivc)

    assert_allclose(problem2["data:mission:sizing:ZFW"], 65076.0, atol=1)
    assert_allclose(problem2["data:mission:sizing:fuel"], 8924.0, atol=1)
    assert_allclose(problem2["data:mission:sizing:fuel:unitary"], 0.0642, rtol=1e-3)

    engine = RubberEngine(5, 30, 1500, 100000, 0.95, 35000 * foot, -50)
    directly_computed_flight_point = FlightPoint(
        mach=0.78,
        altitude=35000 * foot,
        engine_setting=EngineSetting.CRUISE,
        thrust=problem["data:propulsion:required_thrust"],
    )
    engine.compute_flight_points(directly_computed_flight_point)
    assert_allclose(
        directly_computed_flight_point.sfc, problem["data:propulsion:SFC"],
    )



#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os.path as pth

import openmdao.api as om
import pytest
from fastoad.io import VariableIO

from tests.testing_utilities import run_system
from ..compute_static_margin import ComputeStaticMargin

DATA_FOLDER_PATH = pth.join(pth.dirname(__file__), "data")
RESULTS_FOLDER_PATH = pth.join(
    pth.dirname(__file__), "results", pth.splitext(pth.basename(__file__))[0]
)


@pytest.fixture(scope="module")
def input_xml() -> VariableIO:
    """
    :return: access to the sample xml data
    """
    # TODO: have more consistency in input data (no need for the whole geometry_inputs_full.xml)
    return VariableIO(pth.join(pth.dirname(__file__), "data", "geometry_inputs_full.xml"))


def test_compute_static_margin(input_xml):
    """ Tests computation of static margin """

    input_vars = om.IndepVarComp()
    input_vars.add_output("data:weight:aircraft:CG:aft:MAC_position", 0.438971)
    input_vars.add_output("data:aerodynamics:cruise:neutral_point:x", 0.537521)

    problem = run_system(ComputeStaticMargin(), input_vars)
    static_margin = problem["data:handling_qualities:static_margin"]
    assert static_margin == pytest.approx(0.098550, abs=1e-6)



"""
Utility functions for OpenMDAO classes/instances
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from copy import deepcopy
from typing import Tuple, List, TypeVar

import numpy as np
import openmdao.api as om


def get_unconnected_input_names(
    problem: om.Problem, promoted_names=False
) -> Tuple[List[str], List[str]]:
    """
    For provided OpenMDAO problem, looks for inputs that are connected to no output.

    .. warning::

        problem.setup() must have been run.

    Inputs that have numpy.nan as default value are considered as mandatory. Other ones are
    considered as optional.

    :param problem: OpenMDAO Problem or System instance to inspect
    :param promoted_names: if True, promoted names will be returned instead of absolute ones
    :return: tuple(list of missing mandatory inputs, list of missing optional inputs)
    """

    model = problem.model

    mandatory_unconnected = set()
    optional_unconnected = set()

    for abs_name, metadata in model.get_io_metadata(
        "input", metadata_keys=["value"], return_rel_names=False
    ).items():
        name = metadata["prom_name"] if promoted_names else abs_name
        if model.get_source(abs_name).startswith("_auto_ivc."):
            if np.all(np.isnan(metadata["value"])):
                mandatory_unconnected.add(name)
            else:
                optional_unconnected.add(name)

    # If a promoted variable is defined both with NaN and non-NaN value, it is
    # considered as mandatory
    if promoted_names:
        optional_unconnected = optional_unconnected - mandatory_unconnected

    return list(mandatory_unconnected), list(optional_unconnected)


T = TypeVar("T", bound=om.Problem)


def get_problem_after_setup(problem: T) -> T:
    """
    Returns a copy of the provided problem, where setup() has been run on the copy.

    .. warning::

        problem.setup() must NOT have been run.

    This method should be used when an operation is needed that requires setup() to be run, without
    having the problem being actually setup.

    :param problem:
    :return: the problem itself if setup() has already been run, or a copy of the provided problem
             after setup() has been run
    """

    tmp_problem = deepcopy(problem)
    tmp_problem.setup()
    return tmp_problem



"""Classes for simulating cruise segments."""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import List

import pandas as pd
from fastoad.base.flight_point import FlightPoint

from .base import RegulatedThrustSegment


class CruiseSegment(RegulatedThrustSegment):
    """
    Class for computing cruise flight segment at constant altitude.

    Mach is considered constant, equal to Mach at starting point.
    Altitude is constant.
    Target is a specified ground_distance. The target definition indicates
    the ground_distance to be covered during the segment, independently of
    the initial value.
    """

    def compute_from(self, start: FlightPoint) -> pd.DataFrame:
        start = FlightPoint(start)
        if start.ground_distance:
            self.target.ground_distance = self.target.ground_distance + start.ground_distance
        return super().compute_from(start)

    def _get_distance_to_target(self, flight_points: List[FlightPoint]) -> bool:
        current = flight_points[-1]
        return self.target.ground_distance - current.ground_distance


class OptimalCruiseSegment(CruiseSegment):
    """
    Class for computing cruise flight segment at maximum lift/drag ratio.

    Mach is considered constant, equal to Mach at starting point. Altitude is set **at every
    point** to get the optimum CL according to current mass.
    """

    def compute_from(self, start: FlightPoint) -> pd.DataFrame:
        start = FlightPoint(start)
        start.altitude = self._get_optimal_altitude(start.mass, start.mach)
        return super().compute_from(start)

    def _compute_next_altitude(self, next_point, previous_point):
        next_point.altitude = self._get_optimal_altitude(
            next_point.mass, previous_point.mach, altitude_guess=previous_point.altitude
        )



#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Union

import numpy as np
import pandas as pd
from numpy.testing import assert_allclose

from fastoad.base.flight_point import FlightPoint
from ..base import AbstractFuelPropulsion, FuelEngineSet


class DummyEngine(AbstractFuelPropulsion):
    def __init__(self, max_thrust, max_sfc):
        """
        Dummy engine model.

        Max thrust does not depend on flight conditions.
        SFC is proportional to thrust rate.

        :param max_thrust: thrust when thrust rate = 1.0
        :param max_sfc: SFC when thrust rate = 1.0
        """
        self.max_thrust = max_thrust
        self.max_sfc = max_sfc

    def compute_flight_points(self, flight_points: Union[FlightPoint, pd.DataFrame]):
        if flight_points.thrust_is_regulated or flight_points.thrust is None:
            flight_points.thrust = self.max_thrust * flight_points.thrust_rate
        else:
            flight_points.thrust_rate = flight_points.thrust / self.max_thrust

        flight_points.sfc = self.max_sfc * flight_points.thrust_rate


def test_EngineSet():
    """Tests behaviour of FuelEngineSet"""
    engine = DummyEngine(1.2e5, 1.0e-5)

    # input = thrust rate
    thrust_rate = np.linspace(0.0, 1.0, 20)
    for engine_count in [1, 2, 3, 4]:
        engine_set = FuelEngineSet(engine, engine_count)
        flight_point = FlightPoint(
            mach=0.0, altitude=0.0, engine_setting=1, thrust_rate=thrust_rate
        )
        engine_set.compute_flight_points(flight_point)
        assert_allclose(flight_point.sfc, engine.max_sfc * thrust_rate)
        assert_allclose(flight_point.thrust, engine.max_thrust * engine_count * thrust_rate)

    # input = thrust
    thrust = np.linspace(0.0, engine.max_thrust, 30)
    for engine_count in [1, 2, 3, 4]:
        engine_set = FuelEngineSet(engine, engine_count)
        flight_point = FlightPoint(mach=0.0, altitude=0.0, engine_setting=1, thrust=thrust)
        engine_set.compute_flight_points(flight_point)
        assert_allclose(flight_point.thrust_rate, thrust / engine_count / engine.max_thrust)
        assert_allclose(flight_point.sfc, engine.max_sfc * flight_point.thrust_rate)



"""
API
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os.path as pth
import sys
import textwrap as tw
from shutil import get_terminal_size
from typing import IO, Union
from time import time

import numpy as np
import openmdao.api as om
import pandas as pd
import requests
from IPython import InteractiveShell
from IPython.display import display, HTML
from whatsopt.show_utils import generate_xdsm_html
from whatsopt.whatsopt_client import WhatsOpt, PROD_URL

from fastoad.cmd.exceptions import FastFileExistsError
from fastoad.io import IVariableIOFormatter
from fastoad.io.configuration import FASTOADProblem
from fastoad.io.configuration.configuration import FASTOADProblemConfigurator
from fastoad.io.xml import VariableLegacy1XmlFormatter
from fastoad.module_management import BundleLoader
from fastoad.module_management import OpenMDAOSystemRegistry
from fastoad.openmdao.variables import VariableList
from fastoad.utils.files import make_parent_dir
from fastoad.utils.postprocessing import OptimizationViewer, VariableViewer
from fastoad.utils.resource_management.copy import copy_resource
from . import resources
from ..module_management.service_registry import RegisterPropulsion

DEFAULT_WOP_URL = "https://ether.onera.fr/whatsopt"
_LOGGER = logging.getLogger(__name__)

SAMPLE_FILENAME = "fastoad.toml"
MAX_TABLE_WIDTH = 200  # For variable list text output


def generate_configuration_file(configuration_file_path: str, overwrite: bool = False):
    """
    Generates a sample configuration file.

    :param configuration_file_path: the path of file to be written
    :param overwrite: if True, the file will be written, even if it already exists
    :raise FastFileExistsError: if overwrite==False and configuration_file_path already exists
    """
    if not overwrite and pth.exists(configuration_file_path):
        raise FastFileExistsError(
            "Configuration file %s not written because it already exists. "
            "Use overwrite=True to bypass." % configuration_file_path,
            configuration_file_path,
        )

    make_parent_dir(configuration_file_path)

    copy_resource(resources, SAMPLE_FILENAME, configuration_file_path)
    _LOGGER.info("Sample configuration written in %s", configuration_file_path)


def generate_inputs(
    configuration_file_path: str,
    source_path: str = None,
    source_path_schema="native",
    overwrite: bool = False,
):
    """
    Generates input file for the :class:`FASTOADProblem` specified in configuration_file_path.

    :param configuration_file_path: where the path of input file to write is set
    :param source_path: path of file data will be taken from
    :param source_path_schema: set to 'legacy' if the source file come from legacy FAST
    :param overwrite: if True, file will be written even if one already exists
    :raise FastFileExistsError: if overwrite==False and configuration_file_path already exists
    """
    conf = FASTOADProblemConfigurator(configuration_file_path)

    input_file_path = conf.input_file_path
    if not overwrite and pth.exists(conf.input_file_path):
        raise FastFileExistsError(
            "Input file %s not written because it already exists. "
            "Use overwrite=True to bypass." % input_file_path,
            input_file_path,
        )

    if source_path_schema == "legacy":
        conf.write_needed_inputs(source_path, VariableLegacy1XmlFormatter())
    else:
        conf.write_needed_inputs(source_path)

    _LOGGER.info("Problem inputs written in %s", input_file_path)


def list_variables(
    configuration_file_path: str,
    out: Union[IO, str] = sys.stdout,
    overwrite: bool = False,
    force_text_output: bool = False,
):
    """
    Writes list of variables for the :class:`FASTOADProblem` specified in configuration_file_path.

    List is generally written as text. It can be displayed as a scrollable table view if:
    - function is used in an interactive IPython shell
    - out == sys.stdout
    - force_text_output == False

    :param configuration_file_path:
    :param out: the output stream or a path for the output file
    :param overwrite: if True and out parameter is a file path, the file will be written even if one
                      already exists
    :param force_text_output: if True, list will be written as text, even if command is used in an
                              interactive IPython shell (Jupyter notebook). Has no effect in other
                              shells or if out parameter is not sys.stdout
    :raise FastFileExistsError: if overwrite==False and out parameter is a file path and the file
                                exists
    """
    conf = FASTOADProblemConfigurator(configuration_file_path)
    problem = conf.get_problem()
    problem.setup()

    # Extracting inputs and outputs
    variables = VariableList.from_problem(problem, get_promoted_names=False)
    variables.sort(key=lambda var: var.name)
    input_variables = VariableList([var for var in variables if var.is_input])
    output_variables = VariableList([var for var in variables if not var.is_input])

    if isinstance(out, str):
        if not overwrite and pth.exists(out):
            raise FastFileExistsError(
                "File %s not written because it already exists. "
                "Use overwrite=True to bypass." % out,
                out,
            )
        make_parent_dir(out)
        out_file = open(out, "w")
        table_width = MAX_TABLE_WIDTH
    else:
        if out == sys.stdout and InteractiveShell.initialized() and not force_text_output:
            # Here we display the variable list as VariableViewer in a notebook
            for var in input_variables:
                var.metadata["I/O"] = "IN"
            for var in output_variables:
                var.metadata["I/O"] = "OUT"

            df = (
                (input_variables + output_variables)
                .to_dataframe()[["I/O", "name", "desc"]]
                .rename(columns={"name": "Name", "desc": "Description"})
            )
            display(HTML(df.to_html()))
            return

        # Here we continue with text output
        out_file = out
        table_width = min(get_terminal_size().columns, MAX_TABLE_WIDTH) - 1

    pd.set_option("display.max_colwidth", 1000)
    max_name_length = np.max(
        [len(name) for name in input_variables.names() + output_variables.names()]
    )
    description_text_width = table_width - max_name_length - 2

    def _write_variables(out_f, variables):
        """Writes variables and their description as a pandas DataFrame"""
        df = variables.to_dataframe()

        # Create a new Series where description are wrapped on several lines if needed.
        # Each line becomes an element of the Series
        df["desc"] = ["\n".join(tw.wrap(s, description_text_width)) for s in df["desc"]]
        new_desc = df.desc.str.split("\n", expand=True).stack()

        # Create a Series for name that will match new_desc Series. Variable name will be in front of
        # first line of description. An empty string will be in front of other lines.
        new_name = [df.name.loc[i] if j == 0 else "" for i, j in new_desc.index]

        # Create the DataFrame that will be displayed
        new_df = pd.DataFrame({"NAME": new_name, "DESCRIPTION": new_desc})

        out_f.write(
            new_df.to_string(
                index=False,
                columns=["NAME", "DESCRIPTION"],
                justify="center",
                formatters={  # Formatters are needed for enforcing left justification
                    "NAME": ("{:%s}" % max_name_length).format,
                    "DESCRIPTION": ("{:%s}" % description_text_width).format,
                },
            )
        )
        out_file.write("\n")

    def _write_text_with_line(txt: str, line_length: int):
        """ Writes a line of given length with provided text inside """
        out_file.write("-" + txt + "-" * (line_length - 1 - len(txt)) + "\n")

    # Inputs
    _write_text_with_line(" INPUTS OF THE PROBLEM ", table_width)
    _write_variables(out_file, input_variables)

    # Outputs
    out_file.write("\n")
    _write_text_with_line(" OUTPUTS OF THE PROBLEM ", table_width)
    _write_variables(out_file, output_variables)
    _write_text_with_line("", table_width)

    if isinstance(out, str):
        out_file.close()
        _LOGGER.info("Output list written in %s", out_file)


def list_systems(
    configuration_file_path: str = None, out: Union[IO, str] = sys.stdout, overwrite: bool = False
):
    """
    Writes list of available systems.
    If configuration_file_path is given and if it defines paths where there are registered systems,
    they will be listed too.

    :param configuration_file_path:
    :param out: the output stream or a path for the output file
    :param overwrite: if True and out is a file path, the file will be written even if one already
                      exists
    :raise FastFileExistsError: if overwrite==False and out is a file path and the file exists
    """

    if configuration_file_path:
        conf = FASTOADProblemConfigurator(configuration_file_path)
        conf.load(configuration_file_path)
    # As the problem has been configured, BundleLoader now knows additional registered systems

    if isinstance(out, str):
        if not overwrite and pth.exists(out):
            raise FastFileExistsError(
                "File %s not written because it already exists. "
                "Use overwrite=True to bypass." % out,
                out,
            )

        make_parent_dir(out)
        out_file = open(out, "w")
    else:
        out_file = out
    out_file.writelines(["== AVAILABLE SYSTEM IDENTIFIERS " + "=" * 68 + "\n", "-" * 100 + "\n"])
    for identifier in sorted(OpenMDAOSystemRegistry.get_system_ids()):
        path = BundleLoader().get_factory_path(identifier)
        domain = OpenMDAOSystemRegistry.get_system_domain(identifier)
        description = OpenMDAOSystemRegistry.get_system_description(identifier)
        if description is None:
            description = ""
        out_file.write("  IDENTIFIER:   %s\n" % identifier)
        out_file.write("  PATH:         %s\n" % path)
        out_file.write("  DOMAIN:       %s\n" % domain.value)
        out_file.write("  DESCRIPTION:  %s\n" % tw.indent(tw.dedent(description), "    "))
        out_file.write("-" * 100 + "\n")
    out_file.write("=" * 100 + "\n")

    out_file.writelines(
        ["\n== AVAILABLE PROPULSION WRAPPER IDENTIFIERS " + "=" * 56 + "\n", "-" * 100 + "\n"]
    )
    for identifier in sorted(RegisterPropulsion.get_model_ids()):
        path = BundleLoader().get_factory_path(identifier)
        description = RegisterPropulsion.get_service_description(identifier)
        if description is None:
            description = ""
        out_file.write("  IDENTIFIER:   %s\n" % identifier)
        out_file.write("  PATH:         %s\n" % path)
        out_file.write("  DESCRIPTION:  %s\n" % tw.indent(tw.dedent(description), "    "))
        out_file.write("-" * 100 + "\n")
    out_file.write("=" * 100 + "\n")

    if isinstance(out, str):
        out_file.close()
        _LOGGER.info("System list written in %s", out_file)


def write_n2(configuration_file_path: str, n2_file_path: str = None, overwrite: bool = False):
    """
    Write the N2 diagram of the problem in file n2.html

    :param configuration_file_path:
    :param n2_file_path:
    :param overwrite:
    """

    if not n2_file_path:
        n2_file_path = pth.join(pth.dirname(configuration_file_path), "n2.html")
    n2_file_path = pth.abspath(n2_file_path)

    if not overwrite and pth.exists(n2_file_path):
        raise FastFileExistsError(
            "N2-diagram file %s not written because it already exists. "
            "Use overwrite=True to bypass." % n2_file_path,
            n2_file_path,
        )

    make_parent_dir(n2_file_path)
    problem = FASTOADProblemConfigurator(configuration_file_path).get_problem()
    problem.setup()
    problem.final_setup()

    om.n2(problem, outfile=n2_file_path, show_browser=False)
    _LOGGER.info("N2 diagram written in %s", n2_file_path)


def write_xdsm(
    configuration_file_path: str,
    xdsm_file_path: str = None,
    overwrite: bool = False,
    depth: int = 2,
    wop_server_url=None,
    api_key=None,
):
    """

    :param configuration_file_path:
    :param xdsm_file_path:
    :param overwrite:
    :param depth:
    :param wop_server_url:
    :param api_key:
    :return:
    """
    if not xdsm_file_path:
        xdsm_file_path = pth.join(pth.dirname(configuration_file_path), "xdsm.html")
    xdsm_file_path = pth.abspath(xdsm_file_path)

    if not overwrite and pth.exists(xdsm_file_path):
        raise FastFileExistsError(
            "XDSM-diagram file %s not written because it already exists. "
            "Use overwrite=True to bypass." % xdsm_file_path,
            xdsm_file_path,
        )

    make_parent_dir(xdsm_file_path)

    problem = FASTOADProblemConfigurator(configuration_file_path).get_problem()
    problem.setup()
    problem.final_setup()

    wop = WhatsOpt(url=wop_server_url, login=False)

    try:
        ok = wop.login(api_key=api_key, echo=False)
    except requests.exceptions.ConnectionError:

        if not wop_server_url and wop.url == PROD_URL:
            used_url = wop.url
            # If connection failed while attempting to reach the wop default URL,
            # that is the internal ONERA server, try with the external server
            try:
                wop = WhatsOpt(url=DEFAULT_WOP_URL)
                ok = wop.login(api_key=api_key, echo=False)
            except requests.exceptions.ConnectionError:
                _LOGGER.warning("Failed to connect to %s and %s", used_url, DEFAULT_WOP_URL)
                return
        else:
            _LOGGER.warning("Failed to connect to %s", wop.url)
            return

    if ok:
        xdsm = wop.push_mda(
            problem, {"--xdsm": True, "--name": None, "--dry-run": False, "--depth": depth}
        )
        generate_xdsm_html(xdsm, xdsm_file_path)
    else:
        wop.logout()
        _LOGGER.warning("Could not login to %s", wop.url)


def _run_problem(
    configuration_file_path: str,
    overwrite: bool = False,
    mode="run_model",
    auto_scaling: bool = False,
) -> FASTOADProblem:
    """
    Runs problem according to provided file

    :param configuration_file_path: problem definition
    :param overwrite: if True, output file will be overwritten
    :param mode: 'run_model' or 'run_driver'
    :param auto_scaling: if True, automatic scaling is performed for design variables and constraints
    :return: the OpenMDAO problem after run
    """

    problem = FASTOADProblemConfigurator(configuration_file_path).get_problem(
        read_inputs=True, auto_scaling=auto_scaling
    )

    outputs_path = pth.normpath(problem.output_file_path)
    if not overwrite and pth.exists(outputs_path):
        raise FastFileExistsError(
            "Problem not run because output file %s already exists. "
            "Use overwrite=True to bypass." % outputs_path,
            outputs_path,
        )

    problem.setup()
    start_time = time()
    if mode == "run_model":
        problem.run_model()
        problem.optim_failed = False  # Actually, we don't know
    else:
        problem.optim_failed = problem.run_driver()
    end_time = time()
    computation_time = round(end_time - start_time, 2)

    problem.write_outputs()
    if problem.optim_failed:
        _LOGGER.error("Optimization failed after " + str(computation_time) + " seconds")
    else:
        _LOGGER.info("Computation finished after " + str(computation_time) + " seconds")

    _LOGGER.info("Problem outputs written in %s", outputs_path)

    return problem


def evaluate_problem(configuration_file_path: str, overwrite: bool = False) -> FASTOADProblem:
    """
    Runs model according to provided problem file

    :param configuration_file_path: problem definition
    :param overwrite: if True, output file will be overwritten
    :return: the OpenMDAO problem after run
    """
    return _run_problem(configuration_file_path, overwrite, "run_model")


def optimize_problem(
    configuration_file_path: str, overwrite: bool = False, auto_scaling: bool = False
) -> FASTOADProblem:
    """
    Runs driver according to provided problem file

    :param configuration_file_path: problem definition
    :param overwrite: if True, output file will be overwritten
    :param auto_scaling: if True, automatic scaling is performed for design variables and constraints
    :return: the OpenMDAO problem after run
    """
    return _run_problem(configuration_file_path, overwrite, "run_driver", auto_scaling=auto_scaling)


def optimization_viewer(configuration_file_path: str):
    """
    Displays optimization information and enables its editing

    :param configuration_file_path: problem definition
    :return: display of the OptimizationViewer
    """
    problem_configuration = FASTOADProblemConfigurator(configuration_file_path)
    viewer = OptimizationViewer()
    viewer.load(problem_configuration)

    return viewer.display()


def variable_viewer(file_path: str, file_formatter: IVariableIOFormatter = None):
    """
    Displays a widget that enables to visualize variables information and edit their values.

    :param file_path: the path of file to interact with
    :param file_formatter: the formatter that defines file format. If not provided, default format will be assumed.
    :return: display of the VariableViewer
    """
    viewer = VariableViewer()
    viewer.load(file_path, file_formatter)

    return viewer.display()



"""Base classes for mission computation."""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from typing import List, Union

import pandas as pd
from fastoad.base.flight_point import FlightPoint
from fastoad.models.performances.mission.polar import Polar
from fastoad.models.propulsion import IPropulsion


class IFlightPart(ABC):
    def compute_from(self, start: FlightPoint) -> pd.DataFrame:
        """
        Computes a flight sequence from provided start point.

        :param start: the initial flight point, defined for `altitude`, `mass` and speed
                      (`true_airspeed`, `equivalent_airspeed` or `mach`). Can also be
                      defined for `time` and/or `ground_distance`.
        :return: a pandas DataFrame where columns names match :attr:`FlightPoint.labels`
        """


class AbstractFlightSequence(IFlightPart):
    """
    Defines and computes a flight sequence.
    """

    def compute_from(self, start: FlightPoint) -> pd.DataFrame:
        segments = []
        segment_start = start
        for segment_calculator in self.flight_sequence:
            flight_points = segment_calculator.compute_from(segment_start)
            if len(segments) > 0:
                # First point of the segment is omitted, as it is the
                # last of previous segment.
                if len(flight_points) > 1:
                    segments.append(flight_points.iloc[1:])
            else:
                # But it is kept if the computed segment is the first one.
                segments.append(flight_points)

            segment_start = FlightPoint(flight_points.iloc[-1])

        if segments:
            return pd.concat(segments).reset_index(drop=True)

    @property
    @abstractmethod
    def flight_sequence(self) -> List[Union[IFlightPart, str]]:
        """
        Defines the sequence as used in :meth:`compute_from`.

        :return: the list of IFlightPart instances for the mission.
        """


class AbstractManualThrustFlightPhase(AbstractFlightSequence, ABC):
    """
    Base class for climb and descent phases.
    """

    def __init__(
        self,
        *,
        propulsion: IPropulsion,
        reference_area: float,
        polar: Polar,
        thrust_rate: float = 1.0,
        name="",
        time_step=None,
    ):
        """
        Initialization is done only with keyword arguments.

        :param propulsion:
        :param reference_area:
        :param polar:
        :param thrust_rate:
        :param time_step: if provided, this time step will be applied for all segments.
        """

        self.segment_kwargs = {
            "propulsion": propulsion,
            "reference_area": reference_area,
            "polar": polar,
            "thrust_rate": thrust_rate,
            "name": name,
            "time_step": time_step,
        }



"""Base classes for flight computation."""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import List, Union

import pandas as pd
from fastoad.base.flight_point import FlightPoint
from scipy.optimize import root_scalar

from ..base import AbstractFlightSequence, IFlightPart
from ..segments.cruise import CruiseSegment


class AbstractSimpleFlight(AbstractFlightSequence):
    """
    Computes a simple flight.

    The computed flight will be be made of:
        - any number of climb phases
        - one cruise segment
        - any number of descent phases.
    """

    def __init__(
        self,
        cruise_distance: float,
        climb_phases: List[AbstractFlightSequence],
        cruise_phase: CruiseSegment,
        descent_phases: List[AbstractFlightSequence],
    ):
        """

        :param cruise_distance:
        :param climb_phases:
        :param cruise_phase:
        :param descent_phases:
        """

        self.climb_phases = climb_phases
        self.cruise_segment = cruise_phase
        self.cruise_segment.target = FlightPoint(ground_distance=cruise_distance, mach="constant")
        self.descent_phases = descent_phases

    @property
    def cruise_distance(self):
        return self.cruise_segment.target

    @cruise_distance.setter
    def cruise_distance(self, cruise_distance):
        self.cruise_segment.target = FlightPoint(ground_distance=cruise_distance)

    @property
    def flight_sequence(self) -> List[Union[IFlightPart, str]]:
        return self.climb_phases + [self.cruise_segment] + self.descent_phases


class RangedFlight(IFlightPart):
    """
    Computes a flight so that it covers the specified distance.
    """

    def __init__(
        self, flight_definition: AbstractSimpleFlight, flight_distance: float,
    ):
        """
        Computes the flight and adjust the cruise distance to achieve the provided flight distance.

        :param flight_definition:
        :param flight_distance: in meters
        """
        self.flight_distance = flight_distance
        self.flight = flight_definition
        self.flight_points = None
        self.distance_accuracy = 0.5e3

    def compute_from(self, start: FlightPoint) -> pd.DataFrame:
        def compute_flight(cruise_distance):
            self.flight.cruise_distance = cruise_distance
            self.flight_points = self.flight.compute_from(start)
            obtained_distance = (
                self.flight_points.iloc[-1].ground_distance
                - self.flight_points.iloc[0].ground_distance
            )
            return self.flight_distance - obtained_distance

        root_scalar(
            compute_flight,
            x0=self.flight_distance * 0.5,
            x1=self.flight_distance * 0.25,
            xtol=0.5e3,
            method="secant",
        )
        return self.flight_points



"""
The base layer for registering and retrieving OpenMDAO systems
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
from types import MethodType
from typing import List, Union, Any

from fastoad.openmdao.types import SystemSubclass
from .bundle_loader import BundleLoader
from .constants import (
    SERVICE_OPENMDAO_SYSTEM,
    OPTION_PROPERTY_NAME,
    DESCRIPTION_PROPERTY_NAME,
    DOMAIN_PROPERTY_NAME,
    ModelDomain,
)
from .exceptions import (
    FastDuplicateFactoryError,
    FastUnknownOMSystemIdentifierError,
    FastDuplicateOMSystemIdentifierException,
    FastBadSystemOptionError,
)

_LOGGER = logging.getLogger(__name__)
"""Logger for this module"""


class OpenMDAOSystemRegistry:
    """
    Class for registering and providing OpenMDAO System objects.
    """

    _loader = BundleLoader()

    @classmethod
    def explore_folder(cls, folder_path: str):
        """
        Explores provided folder for Systems to register (i.e. modules that use
        :meth:`~OpenMDAOSystemRegistry.register_system` )

        :param folder_path:
        """
        cls._loader.install_packages(folder_path)

    @classmethod
    def register_system(
        cls,
        system_class: type,
        identifier: str,
        domain: ModelDomain = None,
        desc=None,
        options: dict = None,
    ):
        """
        Registers the System (or subclass) so it can later be retrieved and
        instantiated.

        :param system_class:
        :param identifier:
        :param domain: information about model domain
        :param desc: description of the model. If not provided, the docstring of
                     the class will be used.
        :param options: options to be transmitted to OpenMDAO class at run-time
        :raise FastDuplicateOMSystemIdentifierException:
        """
        try:
            properties = {
                DOMAIN_PROPERTY_NAME: domain if domain else ModelDomain.UNSPECIFIED,
                DESCRIPTION_PROPERTY_NAME: desc if desc else system_class.__doc__,
                OPTION_PROPERTY_NAME: options if options else {},
            }
            factory = cls._loader.register_factory(
                system_class, identifier, SERVICE_OPENMDAO_SYSTEM, properties
            )
            return factory
        except FastDuplicateFactoryError as err:
            # Just a more specialized error message
            raise FastDuplicateOMSystemIdentifierException(err.factory_name)

    @classmethod
    def get_system_ids(cls, properties: dict = None) -> List[str]:
        """

        :param properties: if provided, only factories that match all provided properties
                           will be returned
        :return: the list of identifiers for registered factories.
        """
        return cls._loader.get_factory_names(SERVICE_OPENMDAO_SYSTEM, properties=properties)

    @classmethod
    def get_system(cls, identifier: str, options: dict = None) -> SystemSubclass:
        """

        :param identifier: identifier of the registered class
        :param options: option values at system instantiation
        :return: an OpenMDAO system instantiated from the registered class
        """

        try:
            properties = cls._loader.get_factory_properties(identifier).copy()
        except ValueError:
            raise FastUnknownOMSystemIdentifierError(identifier)

        if options:
            properties[OPTION_PROPERTY_NAME] = properties[OPTION_PROPERTY_NAME].copy()
            properties[OPTION_PROPERTY_NAME].update(options)

        system = cls._loader.instantiate_component(identifier, properties=properties)

        # Before making the system available to get options from OPTION_PROPERTY_NAME,
        # check that options are valid to avoid failure at setup()
        options = getattr(system, "_" + OPTION_PROPERTY_NAME, None)
        if options:
            invalid_options = [name for name in options if name not in system.options]
            if invalid_options:
                raise FastBadSystemOptionError(identifier, invalid_options)

        decorated_system = _option_decorator(system)
        return decorated_system

    @classmethod
    def get_system_domain(cls, system_or_id: Union[str, SystemSubclass]) -> ModelDomain:
        """

        :param system_or_id: an identifier of a registered OpenMDAO System class or
                             an instance of a registered OpenMDAO System class
        :return: the model domain associated to given system or system identifier
        """
        return cls._get_system_property(system_or_id, DOMAIN_PROPERTY_NAME)

    @classmethod
    def get_system_description(cls, system_or_id: Union[str, SystemSubclass]) -> str:
        """

        :param system_or_id: an identifier of a registered OpenMDAO System class or
                             an instance of a registered OpenMDAO System class
        :return: the description associated to given system or system identifier
        """

        return cls._get_system_property(system_or_id, DESCRIPTION_PROPERTY_NAME)

    @classmethod
    def _get_system_property(
        cls, system_or_id: Union[str, SystemSubclass], property_name: str
    ) -> Any:
        """

        :param system_or_id: an identifier of a registered OpenMDAO System class or
                             an instance of a registered OpenMDAO System class
        :param property_name:
        :return: the property value associated to given system or system identifier
        """

        if isinstance(system_or_id, str):
            return BundleLoader().get_factory_property(system_or_id, property_name)
        else:
            return BundleLoader().get_instance_property(system_or_id, property_name)


def _option_decorator(instance: SystemSubclass) -> SystemSubclass:
    """
    Decorates provided OpenMDAO instance so that instance.options are populated
    using iPOPO property named after OPTION_PROPERTY_NAME constant.

    :param instance: the instance to decorate
    :return: the decorated instance
    """

    # Rationale:
    #   The idea here is to populate the options at `setup()` time while keeping
    #   all the operations that are in the original `setup()` of the class.
    #
    #   This could have been done by making all our OpenMDAO classes inherit from
    #   a base class where the option values are retrieved, but modifying each
    #   OpenMDAO class looks overkill. Moreover, it would add to them a dependency
    #   to FAST-OAD after having avoided to introduce dependencies outside OpenMDAO.
    #   Last but not least, we would need future contributor to stick to this practice
    #   of inheritance.
    #
    #   Therefore, the most obvious alternative is a decorator. In this decorator, we
    #   could have produced a new instance of the same class that has its own `setup()`
    #   that calls the original `setup()` (i.e. the original Decorator pattern AIUI)
    #   but the new instance would be out of iPOPO's scope.
    #   So we just modify the original instance where we need to "replace"
    #   the `setup()` method to have our code called automagically, without losing the
    #   initial code of `setup()` where there is probably important things. So the trick
    #   is to rename the original `setup()` as `original_setup()`, and create a new
    #   `setup()` that does its job and then calls `original_setup()`.

    def setup(self):
        """ Will replace the original setup() method"""

        # Use values from iPOPO option properties
        option_dict = getattr(self, "_" + OPTION_PROPERTY_NAME, None)
        if option_dict:
            for name, value in option_dict.items():
                self.options[name] = value

        # Call the original setup method
        self.__setup_before_option_decorator()

    # Move the (already bound) method "setup" to "__setup_before_option_decorator"
    setattr(instance, "__setup_before_option_decorator", instance.setup)

    # Create and bind the new "setup" method
    setup_method = MethodType(setup, instance)
    setattr(instance, "setup", setup_method)

    return instance



#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from copy import deepcopy

import openmdao.api as om

from fastoad.io import IVariableIOFormatter, VariableIO
from fastoad.openmdao.validity_checker import ValidityDomainChecker
from fastoad.openmdao.variables import VariableList

INPUT_SYSTEM_NAME = "inputs"


class FASTOADProblem(om.Problem):
    """Vanilla OpenMDAO Problem except that it can read and write its variables from/to files.

    It also runs :class:`~fastoad.openmdao.validity_checker.ValidityDomainChecker`
    after each :meth:`run_model` or :meth:`run_driver`
    (but it does nothing if no check has been registered)

    A classical usage of this class would be::

        problem = FASTOADProblem()  # instantiation
        [... configuration as for any OpenMDAO problem ...]

        problem.input_file_path = "inputs.xml"
        problem.output_file_path = "outputs.xml"
        problem.write_needed_inputs()  # writes the input file (defined above) with
                                       # needed variables so user can fill it with proper values
        # or
        problem.write_needed_inputs('previous.xml')  # writes the input file with needed variables
                                                     # and values taken from provided file when
                                                     # available
        problem.read_inputs()    # reads the input file
        problem.run_driver()     # runs the OpenMDAO problem
        problem.write_outputs()  # writes the output file
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.input_file_path = None
        self.output_file_path = None

    def run_model(self, case_prefix=None, reset_iter_counts=True):
        status = super().run_model(case_prefix, reset_iter_counts)
        ValidityDomainChecker.check_problem_variables(self)
        return status

    def run_driver(self, case_prefix=None, reset_iter_counts=True):
        status = super().run_driver(case_prefix, reset_iter_counts)
        ValidityDomainChecker.check_problem_variables(self)
        return status

    def write_needed_inputs(
        self, source_file_path: str = None, source_formatter: IVariableIOFormatter = None,
    ):
        """
        Writes the input file of the problem with unconnected inputs of the problem.

        .. warning::

            :meth:`setup` must have been run on this Problem instance.

        Written value of each variable will be taken:
        1. from input_data if it contains the variable
        2. from defined default values in component definitions

        WARNING: if inputs have already been read, they won't be needed any more
        and won't be in written file.

        :param source_file_path: if provided, variable values will be read from it
        :param source_formatter: the class that defines format of input file. if not provided,
                                expected format will be the default one.
        """
        variables = VariableList.from_unconnected_inputs(self, with_optional_inputs=True)
        if source_file_path:
            ref_vars = VariableIO(source_file_path, source_formatter).read()
            variables.update(ref_vars)
            for var in variables:
                var.is_input = True
        writer = VariableIO(self.input_file_path)
        writer.write(variables)

    def read_inputs(self):
        """
        Reads inputs from the configured input file.
        """
        if self.input_file_path:
            # Reads input file
            reader = VariableIO(self.input_file_path)
            variables = reader.read()

            ivc = variables.to_ivc()

            # ivc will be added through add_subsystem, but we must use set_order() to
            # put it first.
            # So we need order of existing subsystem to provide the full order list to
            # set_order() to get order of systems, we use system_iter() that can be used
            # only after setup().
            # But we will not be allowed to use add_subsystem() after setup().
            # So we use setup() on a copy of current instance, and get order from this copy
            tmp_prob = deepcopy(self)
            tmp_prob.setup()
            previous_order = [
                system.name
                for system in tmp_prob.model.system_iter(recurse=False)
                if system.name != "_auto_ivc"
                # OpenMDAO 3.2+ specific : _auto_ivc is an output of system_iter() but is not
                # accepted as input of set_order()
            ]

            self.model.add_subsystem(INPUT_SYSTEM_NAME, ivc, promotes=["*"])
            self.model.set_order([INPUT_SYSTEM_NAME] + previous_order)

    def write_outputs(self):
        """
        Writes all outputs in the configured output file.
        """
        if self.output_file_path:
            writer = VariableIO(self.output_file_path)
            variables = VariableList.from_problem(self)
            writer.write(variables)



"""Main components for mass breakdown."""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import openmdao.api as om

from fastoad.models.options import PAYLOAD_FROM_NPAX
from .a_airframe import (
    WingWeight,
    FuselageWeight,
    EmpennageWeight,
    FlightControlsWeight,
    LandingGearWeight,
    PylonsWeight,
    PaintWeight,
)
from .b_propulsion import (
    EngineWeight,
    FuelLinesWeight,
    UnconsumablesWeight,
)
from .c_systems import (
    PowerSystemsWeight,
    LifeSupportSystemsWeight,
    NavigationSystemsWeight,
    TransmissionSystemsWeight,
    FixedOperationalSystemsWeight,
    FlightKitWeight,
)
from .cs25 import Loads
from .d_furniture import (
    PassengerSeatsWeight,
    FoodWaterWeight,
    SecurityKitWeight,
    ToiletsWeight,
)
from .e_crew import CrewWeight
from .payload import ComputePayload
from .update_mlw_and_mzfw import UpdateMLWandMZFW


class MTOWComputation(om.AddSubtractComp):
    """
    Computes MTOW from OWE, design payload and consumed fuel in sizing mission.
    """

    def setup(self):
        self.add_equation(
            "data:weight:aircraft:MTOW",
            [
                "data:weight:aircraft:OWE",
                "data:weight:aircraft:payload",
                "data:mission:sizing:fuel",
            ],
            units="kg",
        )


class MassBreakdown(om.Group):
    """
    Computes analytically the mass of each part of the aircraft, and the resulting sum,
    the Overall Weight Empty (OWE).

    Some models depend on MZFW (Max Zero Fuel Weight), MLW (Max Landing Weight) and
    MTOW (Max TakeOff Weight), which depend on OWE.

    This model cycles for having consistent OWE, MZFW and MLW.

    Options:
    - payload_from_npax: If True (default), payload masses will be computed from NPAX, if False
                         design payload mass and maximum payload mass must be provided.
    """

    def initialize(self):
        self.options.declare(PAYLOAD_FROM_NPAX, types=bool, default=True)

    def setup(self):
        if self.options[PAYLOAD_FROM_NPAX]:
            self.add_subsystem("payload", ComputePayload(), promotes=["*"])
        self.add_subsystem("owe", OperatingWeightEmpty(), promotes=["*"])
        self.add_subsystem("update_mzfw_and_mlw", UpdateMLWandMZFW(), promotes=["*"])

        # Solvers setup
        self.nonlinear_solver = om.NonlinearBlockGS()
        self.nonlinear_solver.options["iprint"] = 0
        self.nonlinear_solver.options["maxiter"] = 50

        self.linear_solver = om.LinearBlockGS()
        self.linear_solver.options["iprint"] = 0


class AirframeWeight(om.Group):
    """
    Computes mass of airframe.
    """

    def setup(self):
        # Airframe
        self.add_subsystem("loads", Loads(), promotes=["*"])
        self.add_subsystem("wing_weight", WingWeight(), promotes=["*"])
        self.add_subsystem("fuselage_weight", FuselageWeight(), promotes=["*"])
        self.add_subsystem("empennage_weight", EmpennageWeight(), promotes=["*"])
        self.add_subsystem("flight_controls_weight", FlightControlsWeight(), promotes=["*"])
        self.add_subsystem("landing_gear_weight", LandingGearWeight(), promotes=["*"])
        self.add_subsystem("pylons_weight", PylonsWeight(), promotes=["*"])
        self.add_subsystem("paint_weight", PaintWeight(), promotes=["*"])

        weight_sum = om.AddSubtractComp()
        weight_sum.add_equation(
            "data:weight:airframe:mass",
            [
                "data:weight:airframe:wing:mass",
                "data:weight:airframe:fuselage:mass",
                "data:weight:airframe:horizontal_tail:mass",
                "data:weight:airframe:vertical_tail:mass",
                "data:weight:airframe:flight_controls:mass",
                "data:weight:airframe:landing_gear:main:mass",
                "data:weight:airframe:landing_gear:front:mass",
                "data:weight:airframe:pylon:mass",
                "data:weight:airframe:paint:mass",
            ],
            units="kg",
            desc="Mass of airframe",
        )

        self.add_subsystem(
            "airframe_weight_sum", weight_sum, promotes=["*"],
        )


class PropulsionWeight(om.Group):
    """
    Computes mass of propulsion.
    """

    def setup(self):
        # Engine have to be computed before pylons
        self.add_subsystem("engines_weight", EngineWeight(), promotes=["*"])
        self.add_subsystem("fuel_lines_weight", FuelLinesWeight(), promotes=["*"])
        self.add_subsystem("unconsumables_weight", UnconsumablesWeight(), promotes=["*"])

        weight_sum = om.AddSubtractComp()
        weight_sum.add_equation(
            "data:weight:propulsion:mass",
            [
                "data:weight:propulsion:engine:mass",
                "data:weight:propulsion:fuel_lines:mass",
                "data:weight:propulsion:unconsumables:mass",
            ],
            units="kg",
            desc="Mass of the propulsion system",
        )

        self.add_subsystem(
            "propulsion_weight_sum", weight_sum, promotes=["*"],
        )


class SystemsWeight(om.Group):
    """
    Computes mass of systems.
    """

    def setup(self):
        self.add_subsystem("power_systems_weight", PowerSystemsWeight(), promotes=["*"])
        self.add_subsystem(
            "life_support_systems_weight", LifeSupportSystemsWeight(), promotes=["*"]
        )
        self.add_subsystem("navigation_systems_weight", NavigationSystemsWeight(), promotes=["*"])
        self.add_subsystem(
            "transmission_systems_weight", TransmissionSystemsWeight(), promotes=["*"]
        )
        self.add_subsystem(
            "fixed_operational_systems_weight", FixedOperationalSystemsWeight(), promotes=["*"]
        )
        self.add_subsystem("flight_kit_weight", FlightKitWeight(), promotes=["*"])

        weight_sum = om.AddSubtractComp()
        weight_sum.add_equation(
            "data:weight:systems:mass",
            [
                "data:weight:systems:power:auxiliary_power_unit:mass",
                "data:weight:systems:power:electric_systems:mass",
                "data:weight:systems:power:hydraulic_systems:mass",
                "data:weight:systems:life_support:insulation:mass",
                "data:weight:systems:life_support:air_conditioning:mass",
                "data:weight:systems:life_support:de-icing:mass",
                "data:weight:systems:life_support:cabin_lighting:mass",
                "data:weight:systems:life_support:seats_crew_accommodation:mass",
                "data:weight:systems:life_support:oxygen:mass",
                "data:weight:systems:life_support:safety_equipment:mass",
                "data:weight:systems:navigation:mass",
                "data:weight:systems:transmission:mass",
                "data:weight:systems:operational:radar:mass",
                "data:weight:systems:operational:cargo_hold:mass",
                "data:weight:systems:flight_kit:mass",
            ],
            units="kg",
            desc="Mass of aircraft systems",
        )

        self.add_subsystem(
            "systems_weight_sum", weight_sum, promotes=["*"],
        )


class FurnitureWeight(om.Group):
    """
    Computes mass of furniture.
    """

    def setup(self):
        self.add_subsystem("passenger_seats_weight", PassengerSeatsWeight(), promotes=["*"])
        self.add_subsystem("food_water_weight", FoodWaterWeight(), promotes=["*"])
        self.add_subsystem("security_kit_weight", SecurityKitWeight(), promotes=["*"])
        self.add_subsystem("toilets_weight", ToiletsWeight(), promotes=["*"])

        weight_sum = om.AddSubtractComp()
        weight_sum.add_equation(
            "data:weight:furniture:mass",
            [
                "data:weight:furniture:passenger_seats:mass",
                "data:weight:furniture:food_water:mass",
                "data:weight:furniture:security_kit:mass",
                "data:weight:furniture:toilets:mass",
            ],
            units="kg",
            desc="Mass of aircraft furniture",
        )

        self.add_subsystem(
            "furniture_weight_sum", weight_sum, promotes=["*"],
        )


class OperatingWeightEmpty(om.Group):
    """Operating Empty Weight (OEW) estimation.

    This group aggregates weight from all components of the aircraft.
    """

    def setup(self):
        # Propulsion should be done before airframe, because it drives pylon mass.
        self.add_subsystem("propulsion_weight", PropulsionWeight(), promotes=["*"])
        self.add_subsystem("airframe_weight", AirframeWeight(), promotes=["*"])
        self.add_subsystem("systems_weight", SystemsWeight(), promotes=["*"])
        self.add_subsystem("furniture_weight", FurnitureWeight(), promotes=["*"])
        self.add_subsystem("crew_weight", CrewWeight(), promotes=["*"])

        weight_sum = om.AddSubtractComp()
        weight_sum.add_equation(
            "data:weight:aircraft:OWE",
            [
                "data:weight:airframe:mass",
                "data:weight:propulsion:mass",
                "data:weight:systems:mass",
                "data:weight:furniture:mass",
                "data:weight:crew:mass",
            ],
            units="kg",
            desc="Mass of crew",
        )

        self.add_subsystem("OWE_sum", weight_sum, promotes=["*"])



"""Definition of the standard flight missions."""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Dict, List, Union

from scipy.constants import foot, knot

from fastoad.base.flight_point import FlightPoint
from fastoad.constants import FlightPhase, EngineSetting
from fastoad.models.propulsion import IPropulsion
from .base import AbstractSimpleFlight
from ..base import IFlightPart, AbstractManualThrustFlightPhase
from ..polar import Polar
from ..segments.altitude_change import AltitudeChangeSegment
from ..segments.cruise import CruiseSegment
from ..segments.speed_change import SpeedChangeSegment


class InitialClimbPhase(AbstractManualThrustFlightPhase):
    """
    Preset for initial climb phase.

    - Climbs up to 400ft at constant EAS
    - Accelerates to EAS = 250kt at constant altitude
    - Climbs up to 1500ft at constant EAS
    """

    @property
    def flight_sequence(self) -> List[Union[IFlightPart, str]]:
        return [
            AltitudeChangeSegment(
                target=FlightPoint(equivalent_airspeed="constant", altitude=400.0 * foot),
                engine_setting=EngineSetting.TAKEOFF,
                **self.segment_kwargs,
            ),
            SpeedChangeSegment(
                target=FlightPoint(equivalent_airspeed=250.0 * knot),
                engine_setting=EngineSetting.TAKEOFF,
                **self.segment_kwargs,
            ),
            AltitudeChangeSegment(
                target=FlightPoint(equivalent_airspeed="constant", altitude=1500.0 * foot),
                engine_setting=EngineSetting.TAKEOFF,
                **self.segment_kwargs,
            ),
        ]


class ClimbPhase(AbstractManualThrustFlightPhase):
    """
    Preset for climb phase.

    - Climbs up to 10000ft at constant EAS
    - Accelerates to EAS = 300kt at constant altitude
    - Climbs up to target altitude at constant EAS
    """

    def __init__(self, **kwargs):
        """
        Uses keyword arguments as for :meth:`AbstractManualThrustFlightPhase` with
        these additional keywords:

        :param maximum_mach: Mach number that won't be exceeded during climb
        :param target_altitude: target altitude in meters, can be a float or
                                AltitudeChangeSegment.OPTIMAL_ALTITUDE to target
                                altitude with maximum lift/drag ratio
        """

        if "maximum_mach" in kwargs:
            self.maximum_mach = kwargs.pop("maximum_mach")
        self.target_altitude = kwargs.pop("target_altitude")
        super().__init__(**kwargs)

    @property
    def flight_sequence(self) -> List[Union[IFlightPart, str]]:
        self.segment_kwargs["engine_setting"] = EngineSetting.CLIMB

        return [
            AltitudeChangeSegment(
                target=FlightPoint(equivalent_airspeed="constant", altitude=10000.0 * foot),
                **self.segment_kwargs,
            ),
            SpeedChangeSegment(
                target=FlightPoint(equivalent_airspeed=300.0 * knot), **self.segment_kwargs,
            ),
            AltitudeChangeSegment(
                target=FlightPoint(equivalent_airspeed="constant", altitude=self.target_altitude),
                **self.segment_kwargs,
                maximum_mach=self.maximum_mach,
            ),
        ]


class DescentPhase(AbstractManualThrustFlightPhase):
    """
    Preset for descent phase.

    - Descends down to EAS = 300kt at constant Mach
    - Descends down to 10000ft at constant EAS
    - Decelerates to EAS = 250kt
    - Descends down to target altitude at constant EAS
    """

    def __init__(self, **kwargs):
        """
        Uses keyword arguments as for :meth:`AbstractManualThrustFlightPhase` with
        this additional keyword:

        :param target_altitude: target altitude in meters
        """

        self.target_altitude = kwargs.pop("target_altitude")
        super().__init__(**kwargs)

    @property
    def flight_sequence(self) -> List[Union[IFlightPart, str]]:
        self.segment_kwargs["engine_setting"] = EngineSetting.IDLE
        return [
            AltitudeChangeSegment(
                target=FlightPoint(equivalent_airspeed=300.0 * knot, mach="constant"),
                **self.segment_kwargs,
            ),
            AltitudeChangeSegment(
                target=FlightPoint(altitude=10000.0 * foot, equivalent_airspeed="constant"),
                **self.segment_kwargs,
            ),
            SpeedChangeSegment(
                target=FlightPoint(equivalent_airspeed=250.0 * knot), **self.segment_kwargs,
            ),
            AltitudeChangeSegment(
                target=FlightPoint(altitude=self.target_altitude, equivalent_airspeed="constant"),
                **self.segment_kwargs,
            ),
        ]


class StandardFlight(AbstractSimpleFlight):
    """
    Defines and computes a standard flight mission.

    The flight sequence is:
    - initial climb
    - climb
    - cruise at constant altitude
    - descent
    """

    def __init__(
        self,
        propulsion: IPropulsion,
        reference_area: float,
        low_speed_climb_polar: Polar,
        high_speed_polar: Polar,
        cruise_mach: float,
        thrust_rates: Dict[FlightPhase, float],
        cruise_distance: float = 0.0,
        climb_target_altitude: float = AltitudeChangeSegment.OPTIMAL_FLIGHT_LEVEL,
        descent_target_altitude: float = 1500.0 * foot,
        time_step=None,
    ):
        """

        :param propulsion:
        :param reference_area:
        :param low_speed_climb_polar:
        :param high_speed_polar:
        :param cruise_mach:
        :param thrust_rates:
        :param cruise_distance:
        :param climb_target_altitude: (in m) altitude where cruise will begin. If value is
                                      AltitudeChangeSegment.OPTIMAL_ALTITUDE (default), climb will
                                      stop when maximum lift/drag ratio is achieved. Cruise will go
                                      on at the same altitude.
        :param descent_target_altitude: (in m) altitude where descent will end in meters
                                        Default is 457.2m (1500ft)
        :param time_step: if provided, this time step will be applied for all segments.
        """

        self.flight_phase_kwargs = {
            "propulsion": propulsion,
            "reference_area": reference_area,
            "time_step": time_step,
        }

        self.low_speed_climb_polar = low_speed_climb_polar
        self.high_speed_polar = high_speed_polar
        self.cruise_mach = cruise_mach
        self.thrust_rates = thrust_rates
        self.climb_target_altitude = climb_target_altitude
        self.descent_target_altitude = descent_target_altitude
        self.time_step = time_step

        kwargs = {
            "propulsion": propulsion,
            "reference_area": reference_area,
            "time_step": time_step,
        }

        initial_climb = InitialClimbPhase(
            **kwargs,
            polar=low_speed_climb_polar,
            thrust_rate=1.0,
            name=FlightPhase.INITIAL_CLIMB.value,
        )
        climb = ClimbPhase(
            **kwargs,
            polar=high_speed_polar,
            thrust_rate=thrust_rates[FlightPhase.CLIMB],
            target_altitude=self.climb_target_altitude,
            maximum_mach=self.cruise_mach,
            name=FlightPhase.CLIMB.value,
        )
        cruise = CruiseSegment(
            **kwargs,
            target=FlightPoint(),
            polar=high_speed_polar,
            engine_setting=EngineSetting.CRUISE,
            name=FlightPhase.CRUISE.value,
        )
        descent = DescentPhase(
            **kwargs,
            polar=high_speed_polar,
            thrust_rate=thrust_rates[FlightPhase.DESCENT],
            target_altitude=self.descent_target_altitude,
            name=FlightPhase.DESCENT.value,
        )
        super().__init__(
            cruise_distance, [initial_climb, climb], cruise, [descent],
        )



"""
Compressibility drag computation.
"""

#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from openmdao.core.explicitcomponent import ExplicitComponent


class CdCompressibility(ExplicitComponent):
    """
    Computation of drag increment due to compressibility effects.

    Formula from §4.2.4 of :cite:`supaero:2014`. This formula can be used for aircraft
    before year 2000.
    Earlier aircraft have more optimized wing profiles that are expected to limit the
    compressibility drag below 2 drag counts. Until a better model can be provided, the
    variable `tuning:aerodynamics:aircraft:cruise:CD:compressibility:ceiling` allows to
    control the maximum authorized compressibility drag.
    """

    def setup(self):

        self.add_input("data:TLAR:cruise_mach", val=np.nan)
        self.add_input("data:aerodynamics:aircraft:cruise:CL", shape_by_conn=True, val=np.nan)
        self.add_input("data:geometry:wing:sweep_25", units="deg", val=np.nan)
        self.add_input("data:geometry:wing:thickness_ratio", val=np.nan)
        self.add_input("tuning:aerodynamics:aircraft:cruise:CD:compressibility:ceiling", val=0.01)
        self.add_output(
            "data:aerodynamics:aircraft:cruise:CD:compressibility",
            copy_shape="data:aerodynamics:aircraft:cruise:CL",
        )

        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs):
        cl = inputs["data:aerodynamics:aircraft:cruise:CL"]
        m = inputs["data:TLAR:cruise_mach"]
        max_cd_comp = inputs["tuning:aerodynamics:aircraft:cruise:CD:compressibility:ceiling"]
        sweep_angle = inputs["data:geometry:wing:sweep_25"]
        thickness_ratio = inputs["data:geometry:wing:thickness_ratio"]

        # Computation of characteristic Mach for 28° sweep and 0.12 of relative thickness
        m_charac_comp_0 = -0.5 * np.maximum(0.35, cl) ** 2 + 0.35 * np.maximum(0.35, cl) + 0.765

        # Computation of characteristic Mach for actual sweep angle and relative thickness
        m_charac_comp = (
            m_charac_comp_0 * np.cos(np.radians(28)) + 0.12 - thickness_ratio
        ) / np.cos(np.radians(sweep_angle))

        cd_comp = np.minimum(max_cd_comp, 0.002 * np.exp(42.58 * (m - m_charac_comp)))

        outputs["data:aerodynamics:aircraft:cruise:CD:compressibility"] = cd_comp



"""Exceptions for module_management package."""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from fastoad.exceptions import FastError


class FastDuplicateFactoryError(FastError):
    def __init__(self, factory_name):
        """
        Raised when trying to register a factory with an already used name.

        :param factory_name:
        """
        super().__init__('Name "%s" is already used.' % factory_name)
        self.factory_name = factory_name


class FastDuplicateOMSystemIdentifierException(FastDuplicateFactoryError):
    """
    Raised when trying to register an OpenMDAO System with an already used identifier.
    """

    def __str__(self):
        """

        :return:
        """
        return (
            "Tried to register an OpenMDAO system with an already used identifier : %s"
            % self.factory_name
        )


class FastNoOMSystemFoundError(FastError):
    def __init__(self, properties):
        """
        Raised when no registered OpenMDAO system could be found from asked properties.

        :param properties:
        """
        super().__init__("No OpenMDAO system found with these properties: %s" % properties)
        self.properties = properties


class FastUnknownOMSystemIdentifierError(FastError):
    def __init__(self, identifier):
        """
        Raised when no OpenMDAO system is registered with asked identifier.

        :param identifier:
        """
        super().__init__("No OpenMDAO system found with this identifier: %s" % identifier)
        self.identifier = identifier


class FastBadSystemOptionError(FastError):
    def __init__(self, identifier, option_names):
        """
        Raised when some option name is not conform to OpenMDAO system definition.

        :param identifier: system identifier
        :param option_names: incorrect option names
        """
        super().__init__(
            "OpenMDAO system %s does not accept following option(s): %s"
            % (identifier, option_names)
        )
        self.identifier = identifier
        self.option_names = option_names


class FastIncompatibleServiceClass(FastError):
    """"""

    def __init__(self, registered_class: type, service_id: str, base_class: type):
        """
        Raised when trying to register as service a class that does not implement
        the specified interface.

        :param registered_class:
        :param service_id:
        :param base_class: the unmatched interface
        """
        super().__init__(
            'Trying to register %s as service "%s" but it does not inherit from %s'
            % (str(registered_class), service_id, str(base_class))
        )
        self.registered_class = registered_class
        self.service_id = service_id
        self.base_class = base_class
